"""This module implements tools to handle dictionaries"""

import sys
from typing import Any, Mapping

import numpy as np

from .array import reshape_to_pdf_shape


def get_val_or_default(in_dict: dict, key: str) -> Any | None:
    """Helper functions to return either an item in a dictionary or the default value of the dictionary

    Parameters
    ----------
    in_dict : dict
        input dictionary
    key : str
        key to search for

    Returns
    -------
    out : Any | None
        The requested item

    Notes
    -----
    This will first try to return:
        in_dict[key] : i.e., the requested item.
    If that fails it will try
        in_dict[None] : i.e., the default for that dictionary.
    If that fails it will return
        None
    """
    if key in in_dict:
        return in_dict[key]
    if None in in_dict:
        return in_dict[None]
    return None


def set_val_or_default(in_dict: dict, key: str, val: Any):
    """Helper functions to either get and item from or add an item to a dictionary and return that item

    Parameters
    ----------
    in_dict : dict
        input dictionary
    key : str
        key to search for
    val : Any
        item to add to the dictionary

    Returns
    -------
    out : Any
        If the key already existed, return the current value. Otherwise, return `val`.

    Notes
    -----
    This will first try to return:
      in_dict[key] : i.e., the requested item.
    If that fails it will return
      `val`
    """
    if key in in_dict:
        return in_dict[key]
    in_dict[key] = val
    return val


def pretty_print(in_dict: dict, prefixes: list, idx=0, stream=sys.stdout) -> None:
    """Print a level of the converstion dictionary in a human-readable format

    Parameters
    ----------
    in_dict : dict
        input dictionary
    prefixs : list
        The prefixs to use at each level of the printing
    idx : int
        The level of the input dictionary we are currently printing
    stream : `stream`
        The stream to print to
    """
    prefix = prefixes[idx]
    for key, val in in_dict.items():
        if key is None:
            key_str = "default"
        else:
            key_str = key
        if isinstance(val, dict):  # pragma: no cover
            stream.write("%s%s:\n" % (prefix, key_str))
            pretty_print(val, prefixes, idx + 1, stream)
        else:
            stream.write("%s%s : %s\n" % (prefix, key_str, val))


def print_dict_shape(in_dict: dict) -> None:
    """Print the shape of arrays in a dictionary.
    This is useful for debugging table creation.

    Parameters
    ----------
    in_dict : dict
        The dictionary to print
    """
    for key, val in in_dict.items():
        print(key, np.shape(val))


def slice_dict(in_dict: dict, subslice: int | slice) -> dict:
    """Create a new dict by taking a slice of of every array in a dict

    Parameters
    ----------
    in_dict : dict
        The dictionary to conver
    subslice : int or slice
        Used to slice the arrays

    Returns
    -------
    out_dict : dict
        The converted dictionary
    """

    out_dict = {}
    for key, val in in_dict.items():
        try:
            out_dict[key] = val[subslice]
        except (KeyError, TypeError):
            out_dict[key] = val
    return out_dict


def check_keys(in_dicts: list[dict]) -> None:
    """Check that the keys in all the in_dicts match

    Raises KeyError if one does not match.
    """
    if not in_dicts:  # pragma: no cover
        return
    master_keys = in_dicts[0].keys()
    for in_dict in in_dicts[1:]:
        if in_dict.keys() != master_keys:  # pragma: no cover
            raise ValueError(
                "Keys to not match: %s != %s" % (in_dict.keys(), master_keys)
            )


def concatenate_dicts(in_dicts: list[dict], add_axis: int = 0) -> dict:
    """Create a new dict by concatenate each array in `in_dicts`

    Parameters
    ----------
    in_dicts : list[dict]
        The dictionaries to stack

    add_axis: int
        The axis to add when ensuring an array is 2D

    Returns
    -------
    out_dict : dict
        The stacked dicionary
    """
    if not in_dicts:  # pragma: no cover
        return {}
    check_keys(in_dicts)
    out_dict = {key: None for key in in_dicts[0].keys()}
    for key in out_dict.keys():
        out_dict[key] = np.concatenate(
            [ensure_2d_array(in_dict[key], add_axis) for in_dict in in_dicts]
        )
    return out_dict


def check_array_shapes(in_dict: dict, npdf: int) -> None:
    """Check that all the arrays in in_dict match the number of pdfs

    Raises ValueError if one does not match.
    """
    if in_dict is None:
        return
    for key, val in in_dict.items():
        if np.size(val) == 1 and npdf == 1:  # pragma: no cover
            continue
        if np.shape(val)[0] != npdf:  # pragma: no cover
            raise ValueError(
                "First dimension of array %s does not match npdf: %i != %i"
                % (key, np.shape(val)[0], npdf)
            )


def compare_two_dicts(d1: dict, d2: dict) -> bool:
    """Check that all the items in d1 and d2 match

    Returns
    -------
    match : bool
        True if they all match, False otherwise
    """
    if d1.keys() != d2.keys():  # pragma: no cover
        return False
    for k, v in d1.items():
        vv = d2[k]
        try:
            if v != vv:  # pragma: no cover
                return False
        except ValueError:
            if not np.allclose(v, vv):  # pragma: no cover
                return False
    return True


def compare_dicts(in_dicts: list[dict]) -> bool:
    """Check that all the dicts in in_dicts match

    Returns
    -------
    match : bool
        True if they all match, False otherwise
    """
    if not in_dicts:  # pragma: no cover
        return True
    first_dict = in_dicts[0]
    for in_dict in in_dicts[1:]:
        if not compare_two_dicts(first_dict, in_dict):  # pragma: no cover
            return False
    return True


def reduce_arrays_to_1d(in_dict: Mapping) -> Mapping:
    """Checks if any arrays in the dictionary have ndim greater than 1, and
    if the first dimension is equal to 1 it reshapes the array to remove that dimension.

    Parameters
    ----------
    in_dict : Mapping
        A dictionary of array-like objects.

    Returns
    -------
    Mapping
        The updated dictionary.
    """

    for key, value in in_dict.items():
        if np.ndim(value) > 1:
            # if shape of any objdata is (1,n) reshape to (,n)
            if np.shape(value)[0] == 1:
                new_val = np.reshape(value, np.shape(value)[-1:])
                in_dict[key] = new_val

    return in_dict


def make_len_equal(in_dict: Mapping[str, np.ndarray], l_arr: int = 1) -> Mapping:
    """Ensures that all arrays in the dictionary have `shape[0]` of at least l_arr.

    This essentially assures that a dictionary of numpy arrays is a `numpyDict`
    `Table-like` object according to `tables_io` to allow for writing.

    Parameters
    ----------
    in_dict : Mapping
        A dictionary of array-like objects.
    l_arr : int, optional
        The value of shape[0] to ensure, by default 1

    Returns
    -------
    Mapping
        The updated dictionary.
    """

    for key, value in in_dict.items():
        if np.shape(value)[0] > l_arr:
            # add a dimension to axis=0
            in_dict[key] = np.expand_dims(value, 0)

    return in_dict


def expand_dimensions(
    in_dict: Mapping[str, np.ndarray], npdf: int, nvals: int
) -> Mapping:

    for key, value in in_dict.items():
        in_dict[key] = reshape_to_pdf_shape(value, npdf, nvals)

    return in_dict


def ensure_2d_array(arr: np.ndarray, add_axis: int = 0) -> np.ndarray:
    """Makes sure that the input array is at least 2 dimensions, by adding a new axis
    to 1D arrays. By default, the new axis is added as axis=0, so the new array will
    have shape (1, len(arr)).

    Parameters
    ----------
    arr : np.ndarray
        The input array
    add_axis : int
        Where to add the new axis.

    Returns
    -------
    np.ndarray
        Returns the input array if it's 2D, or an array with an extra dimension if
        given a 1D array.

    """

    if np.ndim(arr) == 1:
        # make array 2D by adding a dimension to given
        return np.expand_dims(arr, add_axis)
    elif np.ndim(arr) >= 2:
        # array is already 2D, no changes needed
        return arr
    elif np.ndim(arr) == 0:
        # this should return the given number with the appropriate number of dimensions
        return np.atleast_2d(arr)



"""qp is a library for managing and converting between different representations of distributions"""

import os

# from ...tests.qp import test_funcs
from .version import __version__

from .parameterizations.spline.spline import (
    spline,
    spline_gen,
    spline_from_xy,
    spline_from_samples,
)
from .parameterizations.hist.hist import hist, hist_gen
from .parameterizations.interp.interp import (
    interp,
    interp_gen,
    interp_irregular,
    interp_irregular_gen,
)
from .parameterizations.quant.quant import quant, quant_gen

from .parameterizations.mixmod.mixmod import (
    mixmod,
    mixmod_gen,
)
from .parameterizations.sparse_interp.sparse import sparse, sparse_gen

# from .parameterizations.analytic.scipy_testdata import *
from .parameterizations.packed_interp.packed_interp import (
    packed_interp,
    packed_interp_gen,
)

from .core.ensemble import Ensemble
from .core.factory import (
    instance,
    add_class,
    create,
    read,
    read_metadata,
    convert,
    concatenate,
    iterator,
    data_length,
    from_tables,
    is_qp_file,
    write_dict,
    read_dict,
    stats,
)
from .core.lazy_modules import *


from .utils import array, interpolation, dictionary, conversion

from .parameterizations.packed_interp import packing_utils


from .core import factory



"""Implementation of an Ensemble of distributions."""

from __future__ import annotations
import os
from typing import Mapping, Optional, Union

import h5py
import numpy as np
import tables_io
from tables_io import hdf5
from typing import Mapping, Optional, Union
from numpy.typing import ArrayLike

from ..utils.dictionary import (
    check_array_shapes,
    compare_dicts,
    concatenate_dicts,
    slice_dict,
    reduce_arrays_to_1d,
    make_len_equal,
    expand_dimensions,
)
from ..utils.array import encode_strings, reduce_dimensions
from ..metrics import quick_moment
from ..parameterizations.base import Pdf_gen

# import psutil
# import timeit


class Ensemble:
    """An object comprised of one or more distributions with the same parameterization.


    The Ensemble allows you to perform operations on the group of parameterizations as a whole.
    An Ensemble has three main data components, the last of which is optional:

    1. The metadata: this contains information about the parameterization, and
       the coordinates of the parameterization.
    2. The object data: this contains the data that is unique to each distribution,
       for example the values that correspond to the coordinates.
    3. The ancillary data (optional): this contains data points where there is one data point
       for each distribution in the ensemble. There can be many of these columns or
       arrays in the ancillary data table.


    Parameters
    ----------
    the_class : Pdf_gen subclass
        The class to use to parameterize the distributions
    data : Mapping
        Dictionary with data used to construct the ensemble. The keys required
        vary for different parameterizations.
    ancil : Optional[Mapping]
        Dictionary with ancillary data, by default None
    method : Optional[str]
        The key for the creation method to use, by default None

    Examples
    --------
    >>> import qp
    >>> import numpy as np
    >>> data = {'bins': [0,1,2,3,4,5],
    ...         'pdfs': np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
    >>> ancil = {'ids': [105, 108]}}
    >>> ens = qp.Ensemble(qp.hist,data,ancil)
    >>> ens.metadata
    {'pdf_name': array([b'hist'], dtype='|S4'),
    'pdf_version': array([0]),
    'bins': array([[0, 1, 2, 3, 4, 5]])}

    """

    def __init__(
        self,
        the_class: Pdf_gen,
        data: Mapping,
        ancil: Optional[Mapping] = None,
        method: Optional[str] = None,
    ):
        """Class constructor. The requirements are the class object that the ensemble is
        to be parameterized as, and the data dictionary.

        The data dictionary will need different keys depending on what parameterization
        you have chosen. If you are unsure which keys are required, try
        ``qp.[parameterization].create_ensemble?``, where [parameterization] is the
        class of ensemble you wish to create. This will output a docstring with which
        describes the necessary inputs (and this function can also be used to create an
        ensemble instead).

        An ancillary data dictionary can also be provided upon creation. This dictionary
        should contain arrays that are the same length as the number of distributions in the
        ensemble. Essentially, this should include arrays of data where each value in the array
        corresponds to a distribution.


        Parameters
        ----------
        the_class : Pdf_gen subclass
            The class to use to parameterize the distributions
        data : Mapping
            Dictionary with data used to construct the ensemble. The keys required
            vary for different parameterizations.
        ancil : Optional[Mapping]
            Dictionary with ancillary data, by default None
        method : Optional[str]
            The key for the creation method to use, by default None

        """
        # start_time = timeit.default_timer()
        self._gen_func = the_class.creation_method(method)
        self._frozen = self._gen_func(**data)
        self._gen_obj = self._frozen.dist
        self._gen_class = type(self._gen_obj)

        self._ancil = None
        self.set_ancil(ancil)

        self._gridded = None
        self._samples = None

    def __repr__(self) -> str:
        class_name = type(self).__name__
        return f"{class_name}(the_class={self._gen_class.name},shape={self.shape})"

    def __len__(self) -> int:
        return self.npdf

    def __getitem__(self, key: Union[int, slice]) -> Ensemble:
        """Build an Ensemble object for a sub-set of the distributions in this ensemble

        Parameters
        ----------
        key : Union [int , slice]
            Used to slice the data to pick out one distribution from this ensemble

        Returns
        -------
        ens : Ensemble
            The ensemble for the requested distribution or slice of distributions
        """
        red_data = {}
        md = self.metadata
        md.pop("pdf_name")
        md.pop("pdf_version")
        for k, v in md.items():
            red_data[k] = np.squeeze(v)

        if self.npdf > 1:
            dd = slice_dict(self.objdata, key)
        elif self.npdf == 1 and key == 0:
            dd = self.objdata
        else:
            raise IndexError(
                f"Cannot slice Ensemble object with {self.npdf} with given index/slice {key}."
            )

        for k, v in dd.items():
            if len(np.shape(v)) < 2:
                red_data[k] = np.expand_dims(v, 0)
            else:
                red_data[k] = v
        if self._ancil is not None and self.npdf > 1:
            ancil = slice_dict(self._ancil, key)
        elif self._ancil is not None and self.npdf == 1:
            ancil = self._ancil
        else:
            ancil = None
        return Ensemble(self._gen_obj, data=red_data, ancil=ancil)

    @property
    def gen_func(self):
        """Return the function used to create the distribution object for this ensemble"""
        return self._gen_func

    @property
    def gen_class(self):
        """Return the class used to generate distributions for this ensemble"""
        return self._gen_class

    @property
    def dist(self):
        """Return the `scipy.stats.rv_continuous` object that generates distributions for this ensemble"""
        return self._gen_obj

    @property
    def kwds(self):
        """Return the kwds associated to the frozen object for this ensemble"""
        return self._frozen.kwds

    @property
    def gen_obj(self):
        """Return the `scipy.stats.rv_continuous` object that generates distributions for this ensemble"""
        return self._gen_obj

    @property
    def frozen(self):
        """Return the `scipy.stats.rv_frozen` object that encapsulates the distributions for this ensemble"""
        return self._frozen

    @property
    def ndim(self) -> int:
        """Return the number of dimensions of distributions in this ensemble."""
        return self._frozen.ndim

    @property
    def shape(self) -> tuple:
        """Return the shape of distributions in this ensemble."""
        return self._frozen.shape

    @property
    def npdf(self) -> int:
        """Return the number of distributions in this ensemble."""
        return self._frozen.npdf

    @property
    def ancil(self) -> Mapping:
        """Return the ancillary data dictionary for this ensemble."""
        return self._ancil

    def x_samples(
        self, min: float = 0.0, max: float = 5.0, n: Optional[int] = 1000
    ) -> np.ndarray[float]:
        """Return an array of x values that can be used to plot all the distributions
        in the Ensemble.

        This is meant to plot the characteristic distribution for an Ensemble of
        discrete data. For example, for an ensemble of histograms that would be
        the PDF, and for an ensemble of quantiles that would be the CDF.

        Analytic parameterizations like `mixmod <qp.mixmod_gen>` or `scipy.stats.norm` will just return a
        `np.linspace(min,max,n) <numpy.linspace>`, and it's recommended you input the values as
        the defaults are the same for all analytic distributions.

        Parameters
        ----------
        min : float, optional
            The minimum x value to be used if the parameterization doesn't have an
            `x_samples` method or is analytic, by default 0.
        max : float, optional
            The maximum x value to be used if the parameterization doesn't have an
            `x_samples` method or is analytic, by default 5.
        n : Optional[int], optional
            The number of points to be used if the parameterization doesn't have an
            `x_samples` method or is analytic, by default 1000

        Returns
        -------
        xs : np.ndarray[float]
            The array of points to use.
        """
        try:
            return self._frozen.dist.x_samples()
        except:
            return np.linspace(min, max, n)

    def convert_to(self, to_class: Pdf_gen, **kwargs: str) -> Ensemble:
        """Convert this ensemble to the given parameterization class. To see
        the available conversion methods for the your chosen parameterization
        and their required arguments, check the docstrings for ``qp.to_class``.
        If the parameterization class doesn't have a conversion methods table,
        then it will not be possible to convert to that class.

        Parameters
        ----------
        to_class :  Pdf_gen subclass
            Parameterization class to convert to
        **kwargs :
            Keyword arguments that are passed to the output class constructor

        Other Parameters
        ----------------
        method : str
            Optional argument to specify a non-default conversion algorithm

        Returns
        -------
        ens : Ensemble
            Ensemble of distributions of type class_to using the data from this object

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]]))
        >>> ens_i = ens_h.convert_to(qp.interp, xvals=np.linspace(0,5,10))
        >>> ens_i.metadata
        {'pdf_name': array([b'interp'], dtype='|S6'),
        'pdf_version': array([0]),
        'xvals': array([0.        , 0.55555556, 1.11111111, 1.66666667, 2.22222222,
        2.77777778, 3.33333333, 3.88888889, 4.44444444, 5.        ]))}

        """
        kwds = kwargs.copy()
        method = kwds.pop("method", None)
        ctor_func = to_class.creation_method(method)
        class_name = to_class.name
        if ctor_func is None:  # pragma: no cover
            raise KeyError(
                "Class named %s does not have a creation_method named %s"
                % (class_name, method)
            )
        extract_func = to_class.extraction_method(method)
        if extract_func is None:  # pragma: no cover
            raise KeyError(
                "Class named %s does not have a extraction_method named %s"
                % (class_name, method)
            )
        data = extract_func(self, **kwds)
        return Ensemble(to_class, data=data, method=method)

    def update(self, data: Mapping, ancil: Optional[Mapping] = None) -> None:
        """Update the frozen distribution object with the given data, and set
        the ancillary data table with ``ancil`` if given.

        Parameters
        ----------
        data : Mapping
            Dictionary with data used to construct the ensemble, including metadata.
        ancil : Optional[Mapping], optional
            Optional dictionary that contains data for each of the distributions
            in the ensemble, by default None.

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([0,0.1,0.1,0.4,0.2]))
        >>> ens_h.update(data={'bins': np.array([1,2,3,4,5]), 'pdfs': np.array([0.1,0.1,0.4,0.2])})
        >>> ens_h.metadata
        {'pdf_name': array([b'hist'], dtype='|S4'),
        'pdf_version': array([0]),
        'bins': array([[1, 2, 3, 4, 5]])}

        """
        self._frozen = self._gen_func(**data)
        self._gen_obj = self._frozen.dist
        self.set_ancil(ancil)
        self._gridded = None
        self._samples = None

    def update_objdata(self, data: Mapping, ancil: Optional[Mapping] = None) -> None:
        """Updates the objdata in the frozen distribution, and sets
        the ancillary data table if given.

        Parameters
        ----------
        data : Mapping
            Dictionary with the object data that will be used to reconstruct the ensemble
        ancil : Optional[Mapping], optional
            Optional dictionary that contains data for each of the distributions
            in the ensemble, by default None.

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([0,0.1,0.1,0.4,0.2]))
        >>> ens_h.objdata
        {'pdfs': array([0.   , 0.125, 0.125, 0.5  , 0.25 ])}
        >>> ens_h.update_objdata(data={'pdfs': np.array([0.05,0.09,0.2,0.3,0.15])})
        >>> ens_h.objdata
        {'pdfs': array([[0.06329114, 0.11392405, 0.25316456, 0.37974684, 0.18987342]])}

        """
        new_data = {}
        for k, v in self.metadata.items():
            if k in ["pdf_name", "pdf_version"]:
                continue
            new_data[k] = np.squeeze(v)
        new_data.update(self.objdata)
        new_data.update(data)
        self.update(new_data, ancil)

    @property
    def metadata(self) -> Mapping:
        """Return the metadata for this ensemble. Metadata are elements that are
        the same for all the distributions in the ensemble. These include the name
        and version of the distribution generation class

        Returns
        -------
        metadata : Mapping
            The dictionary of the metadata.

        """

        dd = {}
        dd.update(self._gen_obj.metadata)
        return dd

    @property
    def objdata(self) -> Mapping:
        """Return the data for this ensemble. These are the elements that differ
        for each distribution in the ensemble. For example, the data points that
        correspond to each of the coordinates given in the metadata.

        Returns
        -------
        objdata : Mapping
            The object data

        Notes
        -----

        If the distribution normalized the data (which many do by default), this
        will return the normalized data and not the original input data.

        """

        dd = {}
        dd.update(self._frozen.kwds)
        dd.pop("row", None)
        dd.update(self._gen_obj.objdata)

        # if there is only one distribution reshape data as necessary
        if self.npdf == 1:
            dd = reduce_arrays_to_1d(dd)

        return dd

    def set_ancil(self, ancil: Mapping) -> None:
        """Set the ancillary data dictionary. The arrays in this dictionary must have
        one row for each of the distributions, which means that the length of these
        arrays (or the first dimension) must be the same as the number of distributions
        in the ensemble.

        Parameters
        ----------
        ancil : Mapping
            The ancillary data dictionary.

        Raises
        ------
        IndexError
            If the length of the arrays in ancil does not match the number of
            distributions in the Ensemble.

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]]))
        >>> ancil = {'ids': np.array([5,7])}
        >>> ens_h.set_ancil(ancil)
        >>> ens_h.ancil
        {'ids': array([5, 7])}

        """
        check_array_shapes(ancil, self.npdf)
        self._ancil = ancil

    def add_to_ancil(self, to_add: Mapping) -> None:  # pragma: no cover
        """Add additional columns to the ancillary data dictionary. The
        ancil dictionary must already exist. If it does not, use `set_ancil`.

        If any of these columns have the same name as already existing
        ancillary data columns, the new columns will overwrite the old ones.


        Parameters
        ----------
        to_add : Mapping
            The columns to add to the ancillary data dict


        Raises
        ------
        IndexError
            If the length of the arrays in to_add does not match the number of
            distributions in the Ensembles

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ancil = {'ids': np.array([5,7])}
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]]), ancil=ancil)
        >>> ens_h.add_to_ancil({'means':np.array([0.2,0.25])})
        >>> ens_h.ancil
        {'ids': array([5, 7]), 'means': array[0.2,0.25]}


        """
        check_array_shapes(to_add, self.npdf)
        self._ancil.update(to_add)

    def append(self, other_ens: Ensemble) -> None:
        """Append another ensemble to this ensemble. The ensembles must be
        of the same parameterization, or this will not work. They must also
        have the same metadata, so for example if they are both histograms
        they must also have the same bins.

        Both ensembles must have an ancillary data dictionary in order for them
        to be appended to each other. If one ensemble has an ancillary data
        dictionary and the other does not, this will set the ancillary data
        dictionary to `None`.

        Parameters
        ----------
        other_ens : Ensemble
            The ensemble to append to this one.

        Raises
        ------
        KeyError
            Raised if the two ensembles do not have matching metadata.

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_1 = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([0,0.1,0.1,0.4,0.2]))
        >>> ens_2 = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([0.5,0.15,0.25,0.45,0.1]))
        >>> ens_1.append(ens_2)
        >>> ens_1.npdf
        2

        """
        if not compare_dicts([self.metadata, other_ens.metadata]):  # pragma: no cover
            raise KeyError("Metadata does not match, can not append")
        full_objdata = concatenate_dicts([self.objdata, other_ens.objdata])
        if self._ancil is not None and other_ens.ancil is not None:  # pragma: no cover
            full_ancil = concatenate_dicts([self.ancil, other_ens.ancil])
        else:
            full_ancil = None
        self.update_objdata(full_objdata, full_ancil)

    def build_tables(self, encode: bool = False, ext: Optional[str] = None) -> Mapping:
        """Returns a dictionary of dictionaries of numpy arrays for the meta data,
        object data, and the ancillary data (if it exists) for this ensemble.

        Parameters
        ----------
        encode : bool
            If True and `ext` is 'hdf5', will encode any string columns in the `ancil` table,
            by default False.
        ext : str, optional
            If set to 'hdf5' when `encode` is True, will encode any string columns
            in the `ancil` table, by default None.


        Returns
        -------
        data : Mapping, `tables_io.TableDict-like`
            The dictionary with the data. Has the keys: ``meta`` for metadata, ``data``
            for object data, and optionally ``ancil`` for ancillary data.

        """
        meta = make_len_equal(self.metadata)
        dd = dict(meta=meta, data=self.objdata)

        # expand out the objdata to 2D arrays if there's only 1 distribution
        if self.npdf == 1:
            new_objdata = expand_dimensions(dd["data"], self.npdf, self.shape[1])
            dd.update(dict(data=new_objdata))

        if self.ancil is not None:
            # encode any string columns if the file will be hdf5
            if encode == True and ext == "hdf5":
                ancil_tmp = encode_strings(self.ancil)
                dd["ancil"] = ancil_tmp
            else:
                dd["ancil"] = self.ancil
        return dd

    def norm(self):
        """Normalizes the input distribution data if it represents a PDF
        and can be normalized.

        Raises
        ------
        AttributeError
            Raised if the parameterization doesn't have a normalization method.
        """

        # get normalized data values
        try:
            normed = self._gen_obj.normalize()
        except AttributeError as err:
            raise AttributeError(
                "This parameterization does not have a normalization function."
            ) from err
        except RuntimeError as err:
            raise err

        # update ensemble objdata with normalized values
        d_keys = list(self.objdata.keys())

        # add in any data from current ensemble which was unchanged
        for key in d_keys:
            if not key in normed.keys():
                normed[key] = self.objdata[key]

        self.update_objdata(data=normed, ancil=self.ancil)

    def mode(self, grid: ArrayLike) -> ArrayLike:
        """Return the mode of each ensemble distribution, evaluated on the given grid.

        Parameters
        ----------
        grid : ArrayLike
            Grid on which to evaluate distribution

        Returns
        -------
        mode : ArrayLike
            The modes of the distributions evaluated on grid, with shape (npdf, 1)

        """
        new_grid, griddata = self.gridded(grid)
        return np.expand_dims(new_grid[np.argmax(griddata, axis=1)], -1)

    def gridded(self, grid: ArrayLike) -> tuple[ArrayLike, ArrayLike]:
        """Build, cache and return the PDF values at the given grid points.
        If the given grid matches the already cached grid, then this just
        returns the cached value.

        Parameters
        ----------
        grid : ArrayLike
            The grid points to evaluate the PDF at.

        Returns
        -------
        gridded : tuple [ ArrayLike, ArrayLike ]
            (grid, pdf_values)


        """
        if self._gridded is None or not np.array_equal(self._gridded[0], grid):
            self._gridded = (grid, self.pdf(grid))
        return self._gridded

    def write_to(self, filename: str) -> None:
        """Write this ensemble to a file.

        The file type can be any of the those supported by tables_io. File type
        is indicated by the suffix of the file name given. Allowed formats are:
        'hdf5','h5','hf5','hd5','fits','fit','pq','parq','parquet'

        If writing to parquet files, a file will be written for the metadata,
        the object data, and the ancillary data if it exists, where the identifying
        key is added to the filename.

        Parameters
        ----------
        filename : str

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_1 = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([0,0.1,0.1,0.4,0.2]))
        >>> ens_1.write_to("hist-ensemble.hdf5")

        """
        basename, ext = os.path.splitext(filename)
        tables = self.build_tables(encode=True, ext=ext[1:])
        tables_io.write(tables, basename, ext[1:])

    def pdf(self, x: ArrayLike) -> ArrayLike:
        """
        Evaluates the probability density function (PDF) for each of the distributions in the ensemble

        Parameters
        ----------
        x : ArrayLike
            Location(s) at which to evaluate the PDF for each distribution.

        Returns
        -------
        pdf : ArrayLike
            The PDF value(s) at the given location(s).

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.pdf(np.linspace(3,6,6))
        array([[0.5       , 0.5       , 0.25      , 0.25      , 0.        ,
                0.        ],
               [0.37974684, 0.37974684, 0.18987342, 0.18987342, 0.        ,
                0.        ]])

        """

        pdf = self._frozen.pdf(x)

        # reduce dimensionality if possible
        if self.npdf == 1:
            pdf = reduce_dimensions(pdf, x)

        return pdf
        # return self._frozen.pdf(x)

    def logpdf(self, x: ArrayLike) -> ArrayLike:
        """
        Evaluates the log of the probability density function (PDF) for each of the distributions in the ensemble.

        Parameters
        ----------
        x : ArrayLike
            Location(s) at which to do the evaluations

        Returns
        -------
        logpdf : ArrayLike
            The log of the PDF at the given location(s)

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.logpdf(np.linspace(3,6,6))
        array([[-0.69314718, -0.69314718, -1.38629436, -1.38629436,        -inf,
               -inf],
              [-0.96825047, -0.96825047, -1.66139765, -1.66139765,        -inf,
               -inf]])

        """
        logpdf = self._frozen.logpdf(x)

        # reduce dimensionality if possible
        if self.npdf == 1:
            logpdf = reduce_dimensions(logpdf, x)

        return logpdf

    def cdf(self, x: ArrayLike) -> ArrayLike:
        """
        Evaluates the cumulative distribution function (CDF) for each of the distributions in the ensemble.

        Parameters
        ----------
        x : ArrayLike
            Location(s) at which to do the evaluations

        Returns
        -------
        cdf : ArrayLike
            The CDF at the given location(s)

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.cdf(np.linspace(3,6,6))
        array([[0.25      , 0.55      , 0.8       , 0.95      , 1.        ,
                1.        ],
               [0.43037975, 0.65822785, 0.84810127, 0.96202532, 1.        ,
                1.        ]])

        """
        cdf = self._frozen.cdf(x)

        # reduce dimensionality if possible
        if self.npdf == 1:
            cdf = reduce_dimensions(cdf, x)

        return cdf

    def logcdf(self, x: ArrayLike) -> ArrayLike:
        """
        Evaluates the log of the cumulative distribution function (CDF) for each of the distributions in the ensemble.

        Parameters
        ----------
        x : ArrayLike
            Location(s) at which to do the evaluations

        Returns
        -------
        cdf : ArrayLike
            The log of the CDF at the given location(s)

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.logcdf(np.linspace(3,6,6))
        array([[-1.38629436, -0.597837  , -0.22314355, -0.05129329,  0.        ,
                0.        ],
               [-0.84308733, -0.41820413, -0.16475523, -0.03871451,  0.        ,
                0.        ]])

        """
        logcdf = self._frozen.logcdf(x)

        # reduce dimensionality if possible
        if self.npdf == 1:
            logcdf = reduce_dimensions(logcdf, x)

        return logcdf

    def ppf(self, q: ArrayLike) -> ArrayLike:
        """
        Evaluates the percentage point function (PPF) for each of the distributions in the ensemble..

        Parameters
        ----------
        q : ArrayLike
            Location(s) at which to do the evaluations

        Returns
        -------
        ppf : ArrayLike
            The PPF at the given location(s)

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.ppf(0.5)
        array([[3.5       ],
               [3.18333333]])

        """
        ppf = self._frozen.ppf(q)

        # reduce dimensionality if possible
        if self.npdf == 1:
            ppf = reduce_dimensions(ppf, q)
        return ppf

    def sf(self, q: ArrayLike) -> ArrayLike:
        """
        Evaluates the survival fraction (SF) for each of the distributions in the ensemble.

        Parameters
        ----------
        q : ArrayLike
            Location(s) at which to evaluate the distributions

        Returns
        -------
        sf : ArrayLike
            The SF at the given location(s)

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.sf(0.5)
        array([[1.        ],
               [0.96835443]])

        """
        sf = self._frozen.sf(q)

        # reduce dimensionality if possible
        if self.npdf == 1:
            sf = reduce_dimensions(sf, q)

        return sf

    def logsf(self, q: ArrayLike) -> ArrayLike:
        """Evaluates the log of the survival function (SF) for each of the distributions in the ensemble.

        Parameters
        ----------
        q : ArrayLike
            Location(s) at which to evaluate the distributions

        Returns
        -------
        sf : ArrayLike
            The log of the SF at the given location(s)

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.logsf(0.5)
        array([[ 0.        ],
               [-0.03215711]])

        """
        logsf = self._frozen.logsf(q)

        # reduce dimensionality if possible
        if self.npdf == 1:
            logsf = reduce_dimensions(logsf, q)

        return logsf

    def isf(self, q: ArrayLike) -> ArrayLike:
        """
        Evaluates the inverse of the survival fraction (SF) for each of the distributions in the ensemble.

        Parameters
        ----------
        q : ArrayLike
            Location(s) at which to evaluate the distributions

        Returns
        -------
        sf : ArrayLike
            The inverse of the survival fraction at the given location(s)

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.isf(0.5)
        array([[3.5       ],
               [3.18333333]])

        """
        isf = self._frozen.isf(q)

        # reduce dimensionality if possible
        if self.npdf == 1:
            isf = reduce_dimensions(isf, q)

        return isf

    def rvs(
        self,
        size: int = 1,
        random_state: Union[None, int, np.random.Generator] = None,
    ) -> ArrayLike:
        """
        Generate samples from the distributions in this ensemble.

        The returned samples are of shape (npdf, size), where size is the number
        of samples per distribution.

        Parameters
        ----------
        size : int, optional
            Number of samples to return, by default 1.
        random_state : int, numpy.random.Generator, None, optional
            The random state to use. Can be provided with a random seed for consistency. By default None.

        Returns
        -------
        samples : ArrayLike
            The array of samples for each distribution in the ensemble, shape (npdf,size)

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.rvs(size=2)
        array([[3.12956247, 3.72090937],
               [4.96783836, 3.24016123]])


        """
        return self._frozen.rvs(
            size=(self._frozen.npdf, size), random_state=random_state
        )

    def stats(self, moments: str = "mv") -> tuple[ArrayLike, ...]:
        """
        Return some statistics for each of the distributions in this ensemble.

        The moments to be returned are determined by the string given to `moments`,
        where each letter represents a specific moment. The options are:
        "m" = mean, "v" = variance, "s" = (Fisher's) skew, "k" = (Fisher's) kurtosis.

        Parameters
        ----------
        moments : str, optional
            Which moments to include, by default "mv"

        Returns
        -------
        stats : tuple[ArrayLike, ... ]
            A sequence of arrays of the moments requested, where the shape of the arrays is (npdf, 1)

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.stats()
        (array([[3.375     ],
                [3.01898734]]),
         array([[0.859375  ],
                [1.23698125]]))

        """
        return self._frozen.stats(moments=moments)

    def median(self) -> ArrayLike:
        """Return the median for each of the distributions in this ensemble.

        Returns
        -------
        medians : ArrayLike
            The median for each distribution, returns a float if there is only one
            distribution, or the shape of the array is (npdf, 1)

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.median()
        array([[3.5       ],
               [3.18333333]])

        """
        median = self._frozen.median()
        # reduce dimensionality if possible
        if self.npdf == 1:
            median = reduce_dimensions(median, 1)

        return median

    def mean(self) -> ArrayLike:
        """Return the mean for each of the distributions in this ensemble.

        Returns
        -------
        means : ArrayLike
            The mean for each distribution, returns a float if there is only one
            distribution, or the shape of the array is (npdf, 1)

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.mean()
        array([[3.375     ],
               [3.01898734]])

        """
        mean = self._frozen.mean()

        # reduce dimensionality if possible
        if self.npdf == 1:
            mean = reduce_dimensions(mean, 1)

        return mean

    def var(self) -> ArrayLike:
        """Return the variance for each of the distributions in this ensemble.

        Returns
        -------
        variances : ArrayLike
            The variance for each distribution, returns a float if there is only one
            distribution, or the shape of the array is (npdf, 1)

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.var()
        array([[0.859375  ],
               [1.23698125]])

        """
        var = self._frozen.var()

        # reduce dimensionality if possible
        if self.npdf == 1:
            var = reduce_dimensions(var, 1)

        return var

    def std(self) -> ArrayLike:
        """Return the standard deviation for each of the distributions in this ensemble.

        Returns
        -------
        stds : ArrayLike
            The standard deviations for each distribution, the shape of the array is (npdf, 1)

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.std()
        array([[0.92702481],
               [1.11219659]])

        """
        std = self._frozen.std()
        # reduce dimensionality if possible
        if self.npdf == 1:
            std = reduce_dimensions(std, 1)

        return std

    def moment(self, n: int) -> ArrayLike:
        """Return the nth moment for each of the distributions in this ensemble.

        Parameters
        ----------
        n : int
            The order of the moment

        Returns
        -------
        moments : ArrayLike
            The nth moment for each distribution, the shape of the array is (npdf, 1)

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.moment(2)
        array([[12.25      ],
               [10.35126582]])

        """
        moment = self._frozen.moment(n)

        # reduce dimensionality if possible
        if self.npdf == 1:
            moment = reduce_dimensions(moment, 1)

        return moment

    def entropy(self) -> ArrayLike:
        """Return the differential entropy for each of the distributions in this ensemble.

        Returns
        -------
        entropy : ArrayLike
            The entropy for each distribution, the shape of the array is (npdf, 1)


        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.entropy()
        array([[1.21300757],
               [1.45307405]])

        """
        entropy = self._frozen.entropy()

        # reduce dimensionality if possible
        if self.npdf == 1:
            entropy = reduce_dimensions(entropy, 1)

        return entropy

    # def pmf(self, k):
    #    """ Return the kth pmf for this ensemble """
    #    return self._frozen.pmf(k)

    # def logpmf(self, k):
    #    """ Return the log of the kth pmf for this ensemble """
    #    return self._frozen.logpmf(k)

    def interval(self, alpha: ArrayLike) -> tuple[ArrayLike, ...]:
        """
        Return the intervals corresponding to a confidence level of `alpha` for each of the
        distributions in this ensemble.

        Parameters
        ----------
        alpha : ArrayLike
            The array of values to return intervals for. These should be the probability that a random variable will be
            drawn from the returned range. Each value should be in the range [0,1].

        Returns
        -------
        interval :  tuple[ArrayLike, ...]
            A tuple of the arrays containing the intervals for each distribution, where the
            shape of the arrays is (npdf, len(alpha))

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.interval(alpha=[0,0.5,0.9])
        (array([[1.4       , 3.        , 3.5       ],
                [0.79      , 2.2875    , 3.18333333]]),
         array([[3.5       , 4.        , 4.8       ],
                [3.18333333, 3.84166667, 4.73666667]]))

        """
        return self._frozen.interval(alpha)

    def histogramize(self, bins: ArrayLike) -> tuple[ArrayLike]:
        """
        Computes integrated histogram bin values for all distributions in the ensemble.

        Parameters
        ----------
        bins : ArrayLike
            Array of N+1 endpoints of N bins

        Returns
        -------
        histogram: tuple[ArrayLike, ArrayLike]
            The first array in the tuple is the bin edges that were input. The second
            array in the tuple is an (npdf, N) array of the values in the bins.

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ens_h.histogramize(bins=np.array([1,2,3,4,5]))
        (array([1, 2, 3, 4, 5]),
         array([[0.125     , 0.125     , 0.5       , 0.25      ],
                [0.11392405, 0.25316456, 0.37974684, 0.18987342]]))
        """
        return self._frozen.histogramize(bins)

    def integrate(
        self, limits: tuple[Union[float, ArrayLike], Union[float, ArrayLike]]
    ) -> ArrayLike:
        """
        Computes the integral under the probability distribution functions (PDFs) of the distributions in the ensemble
        between the given limits.

        Parameters
        ----------
        limits : tuple[Union[float, ArrayLike], Union[float, ArrayLike]]
            A tuple with the limits of integration, where the first object in the tuple is
            the lower limit, and the second object is the upper limit. The limit objects can
            be floats or arrays, where the number of limits is the length of those arrays, or
            `nlimits`.


        Returns
        -------
        integral: ArrayLike
            Value of the integral(s), with the shape (npdf, nlimits)
        """
        return self.cdf(limits[1]) - self.cdf(limits[0])

    def mix_mod_fit(self, comps=5):  # pragma: no cover
        """
        Fits the parameters of a given functional form to an approximation

        Parameters
        ----------
        comps : int, optional
            Number of components to consider
        using : str, optional
            Which existing approximation to use, defaults to first approximation
        vb : bool
            Report progress

        Returns
        -------
        self.mix_mod: list [ `qp.Composite` ]
            List of `qp.Composite` objects approximating the PDFs

        Notes
        -----
        Currently only supports mixture of Gaussians
        """
        raise NotImplementedError("mix_mod_fit %i" % comps)

    def moment_partial(self, n: int, limits: tuple, dx: float = 0.01) -> ArrayLike:
        """Return the nth moment over a particular range for each of the distributions in this ensemble.

        Parameters
        ----------
        n : int
            The order of the moment to return
        limits : tuple
            The range over which to calculate the moment, where the second number is the
            upper limit.
        dx : float, optional
            The distance between grid points when calculating, by default 0.01

        Returns
        -------
        ArrayLike
            Array of the moments for each of the distributions, with shape (npdf,)

        """
        D = int((limits[-1] - limits[0]) / dx)
        grid = np.linspace(limits[0], limits[1], D)
        # dx = (limits[-1] - limits[0]) / (D - 1)

        P_eval = self.gridded(grid)[1]
        grid_to_n = grid**n
        return quick_moment(P_eval, grid_to_n, dx)

    def plot(
        self,
        key: Union[int, slice] = 0,
        **kwargs: str,
    ):
        """Plot the selected distribution as a curve.

        Parameters
        ----------
        key : int or slice, optional
            The index or slice of the distribution or distributions from this ensemble
            to plot, by default 0.

        Other Parameters
        ----------------
        axes : Axes
            The axes to plot on. Either this or xlim must be provided.
        xlim : (float, float)
            The x-axis limits. Either this or axes must be provided.
        kwargs :
            Any keyword arguments to pass to matplotlib's axes.plot() method.

        Returns
        -------
        axes : Axes
            The plot axes
        """
        return self._gen_class.plot(self[key], **kwargs)

    def plot_native(self, key: Union[int, slice] = 0, **kwargs: str):
        """Plot the selected distribution in the default format for this parameterization. To find what arguments are
        required for specific parameterizations, you can check the docstrings
        of ``qp.[parameterization].plot_native``, where ``[parameterization]`` is the parameterization
        class for the current ensemble.

        Parameters
        ----------
        key : int or slice, optional
            The index or slice of the distribution or distributions from this ensemble
            to plot, by default 0.
        kwargs :
            The keyword arguments to pass to the parameterization's plot_native method.

        Returns
        -------
        axes : Axes
            The plot axes


        """
        return self._gen_class.plot_native(self[key], **kwargs)

    def _get_allocation_kwds(self, npdf: int) -> Mapping:
        tables = self.build_tables()
        keywords = {}
        for group, tab in tables.items():
            if group != "meta":
                keywords[group] = {}
                for key, array in tab.items():
                    shape = list(array.shape)
                    shape[0] = npdf
                    keywords[group][key] = (shape, array.dtype)
        return keywords

    def initializeHdf5Write(
        self, filename: str, npdf: int, comm=None
    ) -> tuple[dict[str, h5py.File | h5py.Group], h5py.File]:
        """Set up the output write for an ensemble, but set size to npdf rather than
        the size of the ensemble, as the "initial chunk" will not contain the full data

        Parameters
        ----------
        filename : str
            Name of the file to create
        npdf : int
            Total number of distributions that the file will contain,
            usually larger then the size of the current ensemble
        comm : MPI communicator
            Optional MPI communicator to allow parallel writing

        Returns
        -------
        group : dict[str, h5py.File | h5py.Group]
            A dictionary of the groups to write to.
        fout : h5py.File
            The output file object that has been created.
        """
        kwds = self._get_allocation_kwds(npdf)
        group, fout = hdf5.initialize_HDF5_write(filename, comm=comm, **kwds)
        return group, fout

    def writeHdf5Chunk(
        self, fname: "h5py.File" | "h5py.Group", start: int, end: int
    ) -> None:
        """Write a chunk of the ensemble data to file. This will write
        the data for the distributions in the slice from [start:end] to the file.
        This includes the ancillary data table.

        Parameters
        ----------
        fname : h5py.File | h5py.Group
            The file or group object to write to
        start : int
            Starting index of data to write in the h5py file
        end : int
            Ending index of data to write in the h5py file
        """
        odict = self.build_tables(encode=True, ext="hdf5").copy()
        odict.pop("meta")
        hdf5.write_dict_to_HDF5_chunk(fname, odict, start, end)

    def finalizeHdf5Write(self, filename: "h5py.File" | "h5py.Group") -> None:
        """Write ensemble metadata to the output file and close the file.

        Parameters
        ----------
        filename : h5py.File | h5py.Group
            The file or group object to complete writing and close.
        """
        mdata = make_len_equal(self.metadata)
        hdf5.finalize_HDF5_write(filename, "meta", **mdata)



"""This module implements a PDT distribution sub-class using interpolated grids"""

from __future__ import annotations
import numpy as np
from scipy.stats import rv_continuous
from typing import Mapping, Optional
from numpy.typing import ArrayLike

from ...core.factory import add_class
from ...core.ensemble import Ensemble
from .packing_utils import PackingType, pack_array, unpack_array
from ..base import Pdf_rows_gen
from ...plotting import get_axes_and_xlims, plot_pdf_on_axes
from ...utils.array import reshape_to_pdf_size
from ...utils.interpolation import interpolate_multi_x_y, interpolate_x_multi_y


def extract_and_pack_vals_at_x(in_dist: "Ensemble", **kwargs):
    """Convert using a set of x and packed y values

    Parameters
    ----------
    in_dist : Ensemble
        Input distributions

    Other Parameters
    ----------------
    xvals : np.ndarray
        Locations at which the pdf is evaluated

    packing_type : PackingType
        Enum specifying the type of packing to use

    Returns
    -------
    data : dict
        The extracted data
    """
    xvals = kwargs.pop("xvals", None)
    packing_type = kwargs.pop("packing_type")
    if xvals is None:  # pragma: no cover
        raise ValueError("To convert to extract_xy_vals you must specify xvals")
    yvals = in_dist.pdf(xvals)
    ypacked, ymax = pack_array(packing_type, yvals, **kwargs)
    return dict(
        xvals=xvals, ypacked=ypacked, ymax=ymax, packing_type=packing_type, **kwargs
    )


class packed_interp_gen(Pdf_rows_gen):  # pylint: disable=too-many-instance-attributes
    """Interpolator based distribution

    Notes
    -----
    This is a version of the interp_pdf that stores the data using a packed integer representation.

    See qp.packing_utils for options on packing

    See qp.interp_pdf for details on interpolation
    """

    # pylint: disable=protected-access

    name = "packed_interp"
    version = 0

    _support_mask = rv_continuous._support_mask

    def __init__(
        self,
        xvals,
        ypacked,
        ymax,
        *args,
        packing_type=PackingType.linear_from_rowmax,
        log_floor=-3.0,
        **kwargs,
    ):
        """
        Create a new distribution by interpolating the given values

        Parameters
        ----------
        xvals : ArrayLike
          The x-values used to do the interpolation
        ypacked : ArrayLike
          The packed version of the y-values used to do the interpolation
        ymax : ArrayLike
          The maximum y-values for each pdf
        packing_type: PackingType
            By default `PackingType.linear_from_rowmax`
        log_floor: float
            By default -3
        """
        if np.size(xvals) != np.shape(ypacked)[-1]:  # pragma: no cover
            raise ValueError(
                "Shape of xbins in xvals (%s) != shape of xbins in yvals (%s)"
                % (np.size(xvals), np.shape(ypacked)[-1])
            )
        self._xvals = np.asarray(xvals)

        # Set support
        self._xmin = self._xvals[0]
        self._xmax = self._xvals[-1]
        # kwargs["shape"] = np.shape(ypacked)[:-1]

        self._yvals = None
        if isinstance(packing_type, PackingType):
            self._packing_type = packing_type.value
        else:
            self._packing_type = packing_type
        self._log_floor = log_floor
        self._ymax = reshape_to_pdf_size(ymax, -1)
        self._ypacked = reshape_to_pdf_size(ypacked, -1)
        kwargs["shape"] = np.shape(self._ypacked)

        check_input = kwargs.pop("check_input", True)
        if check_input:
            self._compute_ycumul()
            self._yvals = (self._yvals.T / self._ycumul[:, -1]).T
            self._ycumul = (self._ycumul.T / self._ycumul[:, -1]).T
        else:  # pragma: no cover
            self._ycumul = None

        super().__init__(*args, **kwargs)
        self._addmetadata("xvals", self._xvals)
        self._addmetadata("packing_type", self._packing_type)
        self._addmetadata("log_floor", self._log_floor)
        self._addobjdata("ypacked", self._ypacked)
        self._addobjdata("ymax", self._ymax)

    def _compute_ycumul(self):
        if self._yvals is None:
            self._unpack()
        copy_shape = np.array(self._yvals.shape)
        self._ycumul = np.ndarray(copy_shape)
        self._ycumul[:, 0] = 0.5 * self._yvals[:, 0] * (self._xvals[1] - self._xvals[0])
        self._ycumul[:, 1:] = np.cumsum(
            (self._xvals[1:] - self._xvals[:-1])
            * 0.5
            * np.add(self._yvals[:, 1:], self._yvals[:, :-1]),
            axis=1,
        )

    def _unpack(self):
        self._yvals = unpack_array(
            PackingType(self._packing_type),
            self._ypacked,
            row_max=self._ymax,
            log_floor=self._log_floor,
        )

    @property
    def xvals(self):
        """Return the x-values used to do the interpolation"""
        return self._xvals

    @property
    def packing_type(self):
        """Returns the packing type"""
        return self._packing_type

    @property
    def log_floor(self):
        """Returns the packing type"""
        return self._log_floor

    @property
    def ypacked(self):
        """Returns the packed y-vals"""
        return self._ypacked

    @property
    def ymax(self):
        """Returns the max for each row"""
        return self._ymax

    @property
    def yvals(self):
        """Return the y-valus used to do the interpolation"""
        return self._yvals

    def _pdf(self, x, row):
        # pylint: disable=arguments-differ
        if self._yvals is None:  # pragma: no cover
            self._unpack()
        pdf = interpolate_x_multi_y(
            x, row, self._xvals, self._yvals, bounds_error=False, fill_value=0.0
        ).ravel()

        # reduce dimension to 0 if there's only one value
        if np.shape(pdf) == (1,) and len(pdf) == 1:
            return pdf[0]
        else:
            return pdf

    def _cdf(self, x, row):
        # pylint: disable=arguments-differ
        if self._ycumul is None:  # pragma: no cover
            self._compute_ycumul()
        return interpolate_x_multi_y(
            x, row, self._xvals, self._ycumul, bounds_error=False, fill_value=(0.0, 1.0)
        ).ravel()

    def _ppf(self, x, row):
        # pylint: disable=arguments-differ
        if self._ycumul is None:  # pragma: no cover
            self._compute_ycumul()

        return interpolate_multi_x_y(
            x,
            row,
            self._ycumul,
            self._xvals,
            bounds_error=False,
            fill_value=(self._xmin, self._xmax),
        ).ravel()

    def _munp(self, m, *args):
        """compute moments"""
        # pylint: disable=arguments-differ
        # Silence floating point warnings from integration.
        with np.errstate(all="ignore"):
            vals = self.custom_generic_moment(m)
        return vals

    def custom_generic_moment(self, m):
        """Compute the mth moment"""
        m = np.asarray(m)
        dx = self._xvals[1] - self._xvals[0]
        if self._yvals is None:  # pragma: no cover
            self._unpack()
        return np.sum(self._xvals**m * self._yvals, axis=1) * dx

    def _updated_ctor_param(self):
        """
        Set the bin edges and packing data as additional constructor argument
        """
        dct = super()._updated_ctor_param()
        dct["xvals"] = self._xvals
        dct["ypacked"] = self._ypacked
        dct["ymax"] = self._ymax
        dct["log_floor"] = self._log_floor
        dct["packing_type"] = PackingType(self._packing_type)
        return dct

    @classmethod
    def get_allocation_kwds(
        cls, npdf, **kwargs
    ) -> dict[str, tuple[tuple[int, int], str]]:
        """Return the keywords necessary to create an 'empty' hdf5 file with npdf entries
        for iterative file writeout.  We only need to allocate the objdata columns, as
        the metadata can be written when we finalize the file.

        Parameters
        ----------
        npdf: int
            number of *total* PDFs that will be written out
        kwargs: dict
            dictionary of kwargs needed to create the ensemble

        Returns
        -------
        dict[str, tuple[tuple[int, int], str]]
        """
        if "xvals" not in kwargs:  # pragma: no cover
            raise ValueError("required argument xvals not included in kwargs")
        ngrid = np.shape(kwargs["xvals"])[-1]
        return dict(
            ypacked=((npdf, ngrid), "u1"),
            ymax=((npdf, 1), "f4"),
        )

    @classmethod
    def plot_native(cls, pdf, **kwargs):
        """Plot the PDF in a way that is particular to this type of distibution

        For a interpolated PDF this uses the interpolation points
        """
        axes, _, kw = get_axes_and_xlims(**kwargs)
        return plot_pdf_on_axes(axes, pdf, pdf.dist.xvals, **kw)

    @classmethod
    def add_mappings(cls):
        """
        Add this classes mappings to the conversion dictionary
        """
        cls._add_creation_method(cls.create, None)
        cls._add_extraction_method(extract_and_pack_vals_at_x, None)

    @classmethod
    def create_ensemble(
        self,
        xvals: ArrayLike,
        ypacked: ArrayLike,
        ymax: ArrayLike,
        packing_type=PackingType.linear_from_rowmax,
        log_floor=-3.0,
        ancil: Optional[Mapping] = None,
    ) -> Ensemble:
        """Creates an Ensemble of distributions parameterized as interpolation that are stored as packed integers.


        Parameters
        ----------
        xvals : ArrayLike
          The x-values used to do the interpolation
        ypacked : ArrayLike
          The packed version of the y-values used to do the interpolation
        ymax : ArrayLike
          The maximum y-values for each pdf
        packing_type: PackingType
            By default `PackingType.linear_from_rowmax`
        log_floor: float
            By default -3

        ancil : Optional[Mapping], optional
            A dictionary of metadata for the distributions, where any arrays have the same length as the number of distributions, by default None

        Returns
        -------
        Ensemble
            An Ensemble object containing all of the given distributions.


        """

        data = {
            "xvals": xvals,
            "ypacked": ypacked,
            "ymax": ymax,
            "packing_type": packing_type,
            "log_floor": log_floor,
        }
        return Ensemble(self, data, ancil)


packed_interp = packed_interp_gen

add_class(packed_interp_gen)



"""This module implements a factory that manages different types of Ensembles"""

from __future__ import annotations
import sys
import os

from collections import OrderedDict
from collections.abc import Iterator
from typing import Mapping, Union, Optional, Tuple

import numpy as np

from scipy import stats as sps

import tables_io
from tables_io import hdf5
from tables_io.types import NUMPY_DICT

from .ensemble import Ensemble

from ..utils.dictionary import compare_dicts, concatenate_dicts, reduce_arrays_to_1d
from ..utils.array import decode_strings

from ..parameterizations.base import Pdf_gen_wrap, Pdf_gen


class Factory(OrderedDict):
    """Factory that creates and manages Ensembles of distributions."""

    def __init__(self):
        """C'tor"""
        super().__init__()
        self._load_scipy_classes()

    @staticmethod
    def _build_data_dict(md_table: Mapping, data_table: Mapping) -> Mapping:
        """Convert the metadata and data tables to a single dictionary that can be used as input to
        build an Ensemble."""
        data_dict = {}

        for col, col_data in md_table.items():
            ndim = np.ndim(col_data)

            if ndim > 1:
                col_data = np.squeeze(col_data)
                if np.ndim(col_data) == 0:
                    col_data = col_data.item()
            elif ndim == 1:
                col_data = col_data[0]

            if isinstance(col_data, bytes):
                col_data = col_data.decode()

            data_dict[col] = col_data

        for col, col_data in data_table.items():
            if len(col_data.shape) < 2:  # pragma: no cover
                data_dict[col] = np.expand_dims(col_data, -1)
            else:
                data_dict[col] = col_data
        return data_dict

    def _make_scipy_wrapped_class(self, class_name: str, scipy_class):
        """Build a qp class from a scipy class"""
        # pylint: disable=protected-access
        override_dict = dict(
            name=class_name,
            version=0,
            freeze=Pdf_gen_wrap._my_freeze,
            _other_init=scipy_class.__init__,
            __doc__=scipy_class.__doc__,
        )
        the_class = type(class_name, (Pdf_gen_wrap, scipy_class), override_dict)

        def create_ensemble(data: Mapping, ancil: Optional[Mapping] = None) -> Ensemble:
            """Creates an Ensemble of distribution(s) in the given parameterization.

            Input data format:
            data = {'arg1': values, 'arg2': values ...} where 'arg1', 'arg2'... are the arguments for the parameterization.
            The length of the values should be the number of distributions being created in the Ensemble, with a minimum value of 1.


            Parameters
            ----------
            data : Mapping
                The dictionary of data for the distributions.
            ancil : Optional[Mapping], optional
                A dictionary of metadata for the distributions, where any arrays have the same length as the number of distributions, by default None

            Returns
            -------
            Ensemble
                An Ensemble object containing all of the given distributions.

            Examples
            --------

            To create an Ensemble with two Gaussian distributions and their associated ids:

            >>> import qp
            >>> data = {'loc': np.array([[0.45],[0.55]]) , 'scale': np.array([[0.2],[0.15]])}
            >>> ancil = {'ids': [20,25]}
            >>> ens = qp.stats.norm.create_ensemble(data,ancil)
            >>> ens.metadata
            {'pdf_name': array([b'norm'], dtype='|S4'), 'pdf_version': array([0])}

            """

            return Ensemble(the_class, data, ancil)

        setattr(
            the_class,
            "create_ensemble",
            create_ensemble,
        )

        self.add_class(the_class)

    def _load_scipy_classes(self):
        """Build qp classes from all the scipy.stats classes"""
        names = sps.__all__
        for name in names:
            attr = getattr(sps, name)
            if isinstance(attr, sps.rv_continuous):
                self._make_scipy_wrapped_class(name, type(attr))

    def add_class(self, the_class: Pdf_gen) -> None:
        """Add a parameterization class to the factory dictionary, so that it is
        included in the set of known parameterization classes. It includes an
        entry both for the actual class name, which ends in ``_gen``, and the
        parameterization name that is also aliased to the class.

        Parameters
        ----------
        the_class : Pdf_gen subclass
            The parameterization class we are adding, which must inherit from `Pdf_gen`.
        """
        # if not isinstance(the_class, Pdf_gen): #pragma: no cover
        #    raise TypeError("Can only add sub-classes of Pdf_Gen to factory")
        if not hasattr(the_class, "name"):  # pragma: no cover
            raise AttributeError(
                "Can not add class %s to factory because it doesn't have a name attribute"
                % the_class
            )
        if the_class.name in self:  # pragma: no cover
            raise KeyError(
                "Class named %s is already in factory, point to %s"
                % (the_class.name, self[the_class.name])
            )
        the_class.add_method_dicts()
        the_class.add_mappings()
        self[the_class.name] = the_class
        setattr(self, "%s_gen" % the_class.name, the_class)
        setattr(self, the_class.name, the_class)

    def create(
        self,
        class_name: Union[str, Pdf_gen],
        data: Mapping,
        method: Optional[str] = None,
        ancil: Optional[Mapping] = None,
    ) -> Ensemble:
        """Make an Ensemble of a particular type of distribution. The ``data``
        dictionary will need different keys depending on what parameterization
        you have chosen.

        If you are unsure which keys are required, try
        ``qp.[parameterization].create_ensemble?``, where [parameterization] is the
        class of ensemble you wish to create. This will output a docstring with the necessary
        inputs (and this function can also be used to create an Ensemble).

        Parameters
        ----------
        class_name : str or class
            The name of the parameterization to make a distribution from.
        data : Mapping
            Dictionary of values passed to the parameterization create function.
        method : str | None, optional
            Used to select which creation method to invoke if there are multiple.
        ancil : Mapping, optional
            Dictionary with ancillary data, by default `None`

        Returns
        -------
        ens : Ensemble
            The newly created Ensemble

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> data = {'bins': [0,1,2,3,4,5],
        ...         'pdfs': np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])}
        >>> ens_h = qp.create('hist', data=data)
        >>> ens.metadata
        {'pdf_name': array([b'hist'], dtype='|S4'),
        'pdf_version': array([0]),
        'bins': array([[0, 1, 2, 3, 4, 5]])}


        """

        # handle if class creation function is given instead of string
        if not isinstance(class_name, str):
            class_name = class_name.name

        if class_name not in self:  # pragma: no cover
            raise KeyError("Class named %s is not in factory" % class_name)
        the_class = self[class_name]
        # ctor_func = the_class.creation_method(method)
        return Ensemble(the_class, data, method=method, ancil=ancil)

    def from_tables(
        self, tables: Mapping, decode: bool = False, ext: str = None
    ) -> Ensemble:
        """Build this Ensemble from a dictionary of tables, where the metadata has key `meta`,
        and the data has key `data`. If there is an ancillary data table, it should have the
        key `ancil`.

        The function will create the ensemble with the parameterization given in the `meta`
        table, and will use any other information in the `meta` table necessary to figure out
        how to construct the ensemble (i.e. construction method).

        Parameters
        ----------
        tables : Mapping
            The dictionary of tables to turn into an Ensemble.
        decode : bool
            If `True` and `ext` is 'hdf5', will decode any string type columns in `ancil`,
            by default False.
        ext : str, optional
            If 'hdf5' and `decode` is True, will decode any string type columns in `ancil`,
            by default None.

        Returns
        -------
        ens : Ensemble
            The ensemble constructed from the data in the tables.

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> meta = {'pdf_name': np.array(['hist'.encode()]), 'pdf_version': np.array([0]),
        ... 'bins':np.array([0,1,2,3,4,5])}
        >>> data = {'pdfs': np.array([[0.  , 0.1 , 0.1 , 0.4 , 0.2 ],
        ... [0.05, 0.09, 0.2 , 0.3 , 0.15]])}
        >>> ens = qp.from_tables({'meta': meta, 'data': data})
        >>> ens.metadata
        {'pdf_name': array([b'hist'], dtype='|S4'),
        'pdf_version': array([0]),
        'bins': array([[0, 1, 2, 3, 4, 5]])}

        """
        md_table = tables["meta"]
        data_table = tables["data"]
        ancil_table = tables.get("ancil")
        if ancil_table is not None:
            if decode == True and ext == "hdf5":
                # decode string arrays
                ancil_table = decode_strings(ancil_table)

        data = self._build_data_dict(md_table, data_table)

        pdf_name = data.pop("pdf_name")
        pdf_version = data.pop("pdf_version")
        if pdf_name not in self:  # pragma: no cover
            raise KeyError("Class named %s is not in factory" % pdf_name)

        # make sure old files are compatible
        if pdf_name == "quant":
            check_input = data.pop("check_input", None)
            if not check_input == None:
                # replace with ensure extent
                data["ensure_extent"] = check_input
        # elif pdf_name == "hist" or pdf_name == "interp":
        #     check_input = data.pop("check_input", None)
        #     if not check_input == None:
        #         # replace with norm
        #         data["norm"] = check_input

        the_class = self[pdf_name]
        reader_convert = the_class.reader_method(pdf_version)
        # ctor_func = the_class.creation_method(None)
        if reader_convert is not None:  # pragma: no cover
            data = reader_convert(data)
        return Ensemble(the_class, data=data, ancil=ancil_table)

    def read_metadata(self, filename: str) -> Mapping:
        """Read an ensemble's metadata from a file, without loading the full data.
        The file must have multiple tables, one of which is called 'meta'.

        Parameters
        ----------
        filename : str
            The full path to the file.

        Returns
        -------
        meta : Mapping
            Returns the metadata table as a dictionary of numpy arrays.

        Examples
        --------

        >>> import qp
        >>> qp.read_metadata("hist-ensemble.hdf5")
        {'pdf_name': array([b'hist'], dtype='|S4'),
        'pdf_version': array([0]),
        'bins': array([[0, 1, 2, 3, 4, 5]])}

        """
        tables = tables_io.read(filename, NUMPY_DICT, keys=["meta"])
        return tables["meta"]

    def is_qp_file(self, filename: str) -> bool:
        """Test if a file is a `qp` file. Must have at least a table called 'meta' in the
        file, and that 'meta' table must have a property 'pdf_name'.

        Parameters
        ----------
        filename : str
            Path to file to test.

        Returns
        -------
        value : bool
            True if the file is a qp file

        Examples
        --------

        >>> import qp
        >>> qp.is_qp_file("test-qpfile.hdf5")
        True

        """
        try:
            # If this isn't a table-like file with a 'meta' table this will throw an exception
            tables = tables_io.read_native(filename, keys=["meta"])
            # If the 'meta' tables doesn't have 'pdf_name' or it is empty this will throw an exception or fail
            return len(tables["meta"]["pdf_name"]) > 0
        except Exception as msg:
            # Any exception means it isn't a qp file
            print(f"This is not a qp file because {msg}")
        return False

    def read(self, filename: str, fmt: Optional[str] = None) -> Ensemble:
        """Read this ensemble from a file. The file must be a `qp` file.

        The function will create the ensemble with the parameterization given in the metadata
        table, and will use any other information in the metadata table necessary to figure out
        how to construct the ensemble (i.e. construction method).

        Parameters
        ----------
        filename : str
            Path to the file.
        fmt : Optional[str], optional
            File format, if `None` it will be taken from the file extension.
            Allowed formats are: 'hdf5','h5','hf5','hd5','fits','fit','pq',
            'parq','parquet'

        Returns
        -------
        ens : Ensemble
            The ensemble constructed from the data in the file.

        Examples
        --------

        >>> import qp
        >>> ens = qp.read("test-qpfile.hdf5")

        """
        _, ext = os.path.splitext(filename)
        if ext in [".pq"]:
            keys = ["data", "meta", "ancil"]
            allow_missing_keys = True
        else:
            keys = None
            allow_missing_keys = False

        tables = tables_io.read(
            filename,
            NUMPY_DICT,
            fmt=fmt,
            keys=keys,
            allow_missing_keys=allow_missing_keys,
        )  # pylint: disable=no-member

        # set up file_fmt to have the file extension information
        if fmt is not None:
            file_fmt = fmt
        else:
            file_fmt = ext[1:]
        return self.from_tables(tables, decode=True, ext=file_fmt)

    def data_length(self, filename: str) -> int:
        """Get the size of data in a file. The file must be a `qp` file, which means
        it must contain an Ensemble with a metadata table.

        Parameters
        ----------
        filename : str
            The path to the file with the data.

        Returns
        -------
        nrows : int
            The length of the data, or the number of distributions in the data.

        Examples
        --------

        >>> import qp
        >>> qp.data_length("hist-ensemble.hdf5")
        2

        """
        f, _ = hdf5.read_HDF5_group(filename, "data")
        num_rows = hdf5.get_group_input_data_length(f)
        return num_rows

    def iterator(
        self,
        filename: str,
        chunk_size: int = 100_000,
        rank: int = 0,
        parallel_size: int = 1,
    ) -> Iterator[int, int, Ensemble]:
        """Iterates through a given Ensemble file and yields a chunk of the ensemble data at a time.
        This means that the returned Ensemble contains the distributions from the returned start
        index to the returned stop index. If there is an ancillary data table, the Ensemble will
        also contain any ancillary data for those distributions.

        Parameters
        ----------
        filename : str
            The path to the file to iterate through.
        chunk_size : int, optional
            The size of chunks to yield, by default 100_000
        rank : int, optional
            The process rank, if run in MPI, by default 0
        parallel_size : int, optional
            The number of processes, if run in MPI, by default 1

        Yields
        ------
        Iterator[int, int, Ensemble]
            the start index, ending index, and an Ensemble with distributions between those two indices

        Raises
        ------
        TypeError
            Raised if this function is run with files that are not ``hdf5`` files.
        KeyError
            Raised if the ``pdf_name`` in the file is not one of the available parameterizations.


        Examples
        --------

        To iterate through an HDF5 Ensemble file, we can use the following code:

        >>> data_file = "./test.hdf5"
        >>> for start, end, ens_chunk in qp.iterator(data_file, chunk_size=11):
        ...     print(f"Indices are: ({start}, {end})")
        ...     print(ens_chunk)
        Indices are: (0, 11)
        Ensemble(the_class=mixmod,shape=(11, 3))
        Indices are: (11, 22)
        Ensemble(the_class=mixmod,shape=(11, 3))
        Indices are: (22, 33)
        Ensemble(the_class=mixmod,shape=(11, 3))
        Indices are: (33, 44)
        Ensemble(the_class=mixmod,shape=(11, 3))
        Indices are: (44, 55)
        Ensemble(the_class=mixmod,shape=(11, 3))
        Indices are: (55, 66)
        Ensemble(the_class=mixmod,shape=(11, 3))
        Indices are: (66, 77)
        Ensemble(the_class=mixmod,shape=(11, 3))
        Indices are: (77, 88)
        Ensemble(the_class=mixmod,shape=(11, 3))
        Indices are: (88, 99)
        Ensemble(the_class=mixmod,shape=(11, 3))
        Indices are: (99, 100)
        Ensemble(the_class=mixmod,shape=(1, 3))


        """
        extension = os.path.splitext(filename)[1]
        if extension not in [".hdf5"]:  # pragma: no cover
            raise TypeError("Can only use qp.iterator on hdf5 files")

        metadata = hdf5.read_HDF5_to_dict(filename, "meta")
        pdf_name = metadata.pop("pdf_name")[0].decode()
        _pdf_version = metadata.pop("pdf_version")[0]
        if pdf_name not in self:  # pragma: no cover
            raise KeyError("Class named %s is not in factory" % pdf_name)
        the_class = self[pdf_name]
        # reader_convert = the_class.reader_method(pdf_version)
        # ctor_func = the_class.creation_method(None)

        f, infp = hdf5.read_HDF5_group(filename, "data")
        try:
            ancil_f, ancil_infp = hdf5.read_HDF5_group(filename, "ancil")
        except KeyError:  # pragma: no cover
            ancil_f, ancil_infp = (None, None)
        num_rows = hdf5.get_group_input_data_length(f)
        ranges = hdf5.data_ranges_by_rank(num_rows, chunk_size, parallel_size, rank)
        data = self._build_data_dict(metadata, {})
        ancil_data = OrderedDict()
        for start, end in ranges:
            for key, val in f.items():
                data[key] = hdf5.read_HDF5_dataset_to_array(val, start, end)
            if ancil_f is not None:
                for key, val in ancil_f.items():
                    ancil_data[key] = hdf5.read_HDF5_dataset_to_array(val, start, end)
            yield start, end, Ensemble(the_class, data=data, ancil=ancil_data)
        infp.close()
        if ancil_infp is not None:
            ancil_infp.close()

    def convert(self, in_dist: Ensemble, class_name: str, **kwds) -> Ensemble:
        """Convert an ensemble to a different parameterization. Keyword arguments are
        required to convert to a different parameterization, but the specific keyword
        arguments required will vary. To check the available conversion methods
        and their associated arguments refer to the docstrings for ``qp.class_name``
        of the parameterization you are converting to. If the class does not
        have a conversion methods table, then it will not be possible to convert
        to that parameterization.


        Parameters
        ----------
        in_dist : Ensemble
            The input Ensemble object to convert.
        class_name : str
            Name of the representation to convert to as a string
        kwds : Mapping
            The arguments required to convert to a function of the given type.

        Returns
        -------
        ens : Ensemble
            The ensemble we converted to

        Examples
        --------

        The following example demonstrates converting from a histogram parameterization
        to an interpolation parameterization. The arguments given will not be the same
        when converting between other parameterizations.

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]]))
        >>> ens_i = qp.convert(ens_h, "interp", xvals=np.linspace(0,5,10))
        >>> ens_i.metadata
        {'pdf_name': array([b'interp'], dtype='|S6'),
        'pdf_version': array([0]),
        'xvals': array([0.        , 0.55555556, 1.11111111, 1.66666667, 2.22222222,
        2.77777778, 3.33333333, 3.88888889, 4.44444444, 5.        ]))}

        """
        kwds_copy = kwds.copy()
        method = kwds_copy.pop("method", None)
        if class_name not in self:  # pragma: no cover
            raise KeyError("Class named %s is not in factory" % class_name)
        if class_name not in self:  # pragma: no cover
            raise KeyError("Class named %s is not in factory" % class_name)
        the_class = self[class_name]
        extract_func = the_class.extraction_method(method)
        if extract_func is None:  # pragma: no cover
            raise KeyError(
                "Class named %s does not have a extraction_method named %s"
                % (class_name, method)
            )
        data = extract_func(in_dist, **kwds_copy)
        return self.create(class_name, data, method)

    def pretty_print(self, stream=sys.stdout) -> None:
        """Print a level of the conversion dictionary in a human-readable format

        Parameters
        ----------
        stream : `stream`
            The stream to print to
        """
        for class_name, cl in self.items():
            stream.write("\n")
            stream.write("%s: %s\n" % (class_name, cl))
            cl.print_method_maps(stream)

    @staticmethod
    def concatenate(ensembles: list[Ensemble]) -> Ensemble:
        """Concatenate a list of Ensembles into one Ensemble. The
        Ensembles must be of the same parameterization and have the same metadata.

        Parameters
        ----------
        ensembles : list[Ensemble]
            The list of ensembles we are concatenating

        Returns
        -------
        ens : Ensemble
            The output

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_1 = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([0,0.1,0.1,0.4,0.2]))
        >>> ens_1.npdf
        1
        >>> ens_2 = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([[0.05,0.09,0.2,0.3,0.15]]))
        >>> ens_2.npdf
        1
        >>> ens_all = qp.concatenate([ens_1, ens_2])
        >>> ens_all.npdf
        2

        """
        if not ensembles:  # pragma: no cover
            return None
        metadata_list = []
        objdata_list = []
        ancil_list = []
        gen_class = None
        for ensemble in ensembles:
            metadata_list.append(ensemble.metadata)
            objdata_list.append(ensemble.objdata)
            if gen_class is None:
                gen_class = ensemble.gen_class
            if ancil_list is not None:
                if ensemble.ancil is None:
                    ancil_list = None
                else:  # pragma: no cover
                    ancil_list.append(ensemble.ancil)
        if not compare_dicts(metadata_list):  # pragma: no cover
            raise ValueError("Metadata does not match")
        metadata = metadata_list[0]
        data = concatenate_dicts(objdata_list)
        if ancil_list is not None:  # pragma: no cover
            ancil = concatenate_dicts(ancil_list, 1)
        else:
            ancil = None
        for k, v in metadata.items():
            if k in ["pdf_name", "pdf_version"]:
                continue
            data[k] = np.squeeze(v)
        return Ensemble(gen_class, data, ancil)

    @staticmethod
    def write_dict(filename: str, ensemble_dict: Mapping[str, Ensemble], **kwargs):
        """Writes out a dictionary of Ensembles to an HDF5 file. Each Ensemble
        in the dictionary will be written to a group, and within each Ensemble group there
        will be subgroups for the metadata, data, and (optional) ancillary data tables.

        Parameters
        ----------
        filename : str
            The file path to write to.
        ensemble_dict : Mapping[str, Ensemble]
            The dictionary of Ensembles to write.
        kwargs :
            Keyword arguments that are passed to the tables_io write_dicts_to_HDF5 function

        Raises
        ------
        ValueError
            Raised if the dictionary contains any values that are not Ensembles.

        Examples
        --------

        >>> import qp
        >>> import numpy as np
        >>> ens_h = qp.hist.create_ensemble(bins= np.array([0,1,2,3,4,5]),
        ... pdfs = np.array([0,0.1,0.1,0.4,0.2]))
        >>> ens_i = qp.interp.create_ensemble(xvals= np.array([0,1,2,3,4]),
        ... yvals = np.array([[0.05,0.09,0.2,0.3,0.15]]))
        >>> qp.write_dict("qp-ensembles.hdf5",{"ens_h": ens_h, "ens_i": ens_i})

        """
        output_tables = {}
        for key, val in ensemble_dict.items():
            # check that val is a qp.Ensemble
            if not isinstance(val, Ensemble):
                raise ValueError(
                    "All values in ensemble_dict must be qp.Ensemble"
                )  # pragma: no cover

            output_tables[key] = val.build_tables(encode=True, ext="hdf5")
        hdf5.write_dicts_to_HDF5(output_tables, filename, **kwargs)

    @staticmethod
    def read_dict(filename: str) -> Mapping[str, Ensemble]:
        """Reads in one or more Ensembles from an HDF5 file to a dictionary of Ensembles.
        The file should contain one top-level group per ensemble. Each Ensemble group should
        have subgroups that are the metadata, data, and (optional) ancillary data tables.

        Parameters
        ----------
        filename : str
            The path to the ``HDF5`` file to read in.

        Returns
        -------
        Mapping[str, Ensemble]
            A dictionary with the Ensembles contained in the file.

        Examples
        --------

        >>> import qp
        >>> ens_dict = qp.read_dict("qp-ensembles.hdf5")

        """
        results = {}

        # retrieve all the top level groups. Assume each top level group
        # corresponds to an ensemble.
        top_level_groups = hdf5.read_HDF5_group_names(filename)

        # for each top level group, convert the subgroups (data, meta, ancil) into
        # a dictionary of dictionaries and pass the result to `from_tables`.
        for top_level_group in top_level_groups:
            tables = {}
            keys = hdf5.read_HDF5_group_names(filename, top_level_group)
            for key_name in keys:
                # retrieve the hdf5 group object
                group_object, _ = hdf5.read_HDF5_group(
                    filename, f"{top_level_group}/{key_name}"
                )

                # use the hdf5 group object to gather data into a dictionary
                tables[key_name] = hdf5.read_HDF5_group_to_dict(group_object)

            results[top_level_group] = from_tables(tables, decode=True, ext="hdf5")

        return results


_FACTORY = Factory()


def instance():
    """Return the factory instance"""
    return _FACTORY


stats = _FACTORY
add_class = _FACTORY.add_class
create = _FACTORY.create
read = _FACTORY.read
read_metadata = _FACTORY.read_metadata
iterator = _FACTORY.iterator
convert = _FACTORY.convert
concatenate = _FACTORY.concatenate
data_length = _FACTORY.data_length
from_tables = _FACTORY.from_tables
is_qp_file = _FACTORY.is_qp_file
write_dict = _FACTORY.write_dict
read_dict = _FACTORY.read_dict



import numpy as np
import pytest

import qp
from tests.helpers.test_funcs import assert_all_close


@pytest.fixture
def quant_ensemble(quant_test_data) -> qp.Ensemble:
    ens = qp.quant.create_ensemble(**quant_test_data["quant"]["ctor_data"])
    return ens


def test_xsamples(quant_ensemble):

    xsamps = quant_ensemble.x_samples()

    xmin = np.min(quant_ensemble.objdata["locs"])
    xmax = np.max(quant_ensemble.objdata["locs"])
    min_dx = np.median(np.diff(quant_ensemble.objdata["locs"]))

    npts = (xmax - xmin) // min_dx
    npts = np.min([int(npts), 10000])
    x_compare = np.linspace(xmin, xmax, npts)

    assert_all_close(xsamps, x_compare)


@pytest.mark.parametrize(
    "quants, locs, match_string",
    [
        (
            np.array([0, 0.5, 1]),
            np.array([0.2, 0.56, np.inf]),
            "There are non-finite values in the locs for the distributions:",
        )
    ],
)
def test_warnings(quants, locs, match_string):
    """Test that warnings when building an interpolated Ensemble work."""

    with pytest.warns(RuntimeWarning, match=match_string):
        ens = qp.quant.create_ensemble(quants=quants, locs=locs)


@pytest.mark.parametrize(
    "quants, locs, match_string",
    [
        (
            np.array([-0.5, 0, 0.5]),
            np.array([0.2, 0.5, 0.6]),
            "Invalid quants: One or more of the given quants is outside the allowed range",
        ),
        (
            np.array([0, 0.5, 0.2]),
            np.array([0.2, 1.0, 1.5]),
            "There are decreasing values, quants must be given in order from 0 to 1",
        ),
        (
            np.array([0, 0.5, 1]),
            np.array([0.5, 0.2, 1.2]),
            "The given data does not produce a one-to-one CDF",
        ),
    ],
)
def test_validate_input(quants, locs, match_string):
    """Tests that the appropriate errors are raised when invalid input is given"""

    with pytest.raises(ValueError, match=match_string) as err:
        ens_h = qp.quant.create_ensemble(quants=quants, locs=locs, warn=False)


@pytest.mark.parametrize(
    "pdf_constructor_name",
    [
        ("piecewise"),
        ("piecewise".encode()),
    ],
)
def test_quant_constructor(pdf_constructor_name):
    """Make sure that quant parameterization doesn't work if the constructor name
    is not in the given list."""

    quants = np.linspace(0, 1, 5)
    locs = np.linspace(-1, 1, 5)

    with pytest.raises(ValueError, match="Unknown interpolator provided:") as exec_info:
        qp.quant.create_ensemble(quants, locs, pdf_constructor_name)



"""This module implements a PDT distribution sub-class using a Gaussian mixture model"""

from __future__ import annotations

import os
import sys
import numpy as np
from scipy.stats import rv_continuous
from scipy import integrate as sciint
from scipy import interpolate as sciinterp
from typing import Mapping, Optional
from numpy.typing import ArrayLike

from . import sparse_rep
from ...core.factory import add_class
from ...core.ensemble import Ensemble
from ..interp.interp import interp_gen
from .sparse_utils import extract_sparse_from_xy


class sparse_gen(interp_gen):
    """Sparse based distribution. The final behavior is similar to interp_gen, but the constructor
    takes a sparse representation to build the interpolator.
    Attempt to inherit from interp_gen : this is failing

    Notes
    -----
    This implements a qp interface to the original code SparsePz from M. Carrasco-Kind.

    """

    # pylint: disable=protected-access

    name = "sparse"
    version = 0

    _support_mask = rv_continuous._support_mask

    def __init__(
        self, xvals, mu, sig, dims, sparse_indices, *args, **kwargs
    ):  # pylint: disable=too-many-arguments
        self.sparse_indices = sparse_indices
        self._xvals = xvals
        self.mu = mu
        self.sig = sig
        self.dims = dims
        cut = kwargs.pop("cut", 1.0e-5)
        # recreate the basis array from the metadata
        sparse_meta = dict(xvals=xvals, mu=mu, sig=sig, dims=dims)
        A = sparse_rep.create_basis(sparse_meta, cut=cut)
        # decode the sparse indices into basis indices and weights
        basis_indices, weights = sparse_rep.decode_sparse_indices(sparse_indices)
        # retrieve the weighted array of basis functions for each object
        pdf_y = A[:, basis_indices] * weights
        # normalize and sum the weighted pdfs
        x = sparse_meta["xvals"]
        y = pdf_y.sum(axis=-1)
        norms = sciint.trapezoid(y.T, x)
        y /= norms
        kwargs.setdefault("xvals", x)
        kwargs.setdefault("yvals", y.T)
        super().__init__(*args, **kwargs)

        self._clearobjdata()
        self._addmetadata("xvals", self._xvals)
        self._addmetadata("mu", self.mu)
        self._addmetadata("sig", self.sig)
        self._addmetadata("dims", self.dims)
        self._addobjdata("sparse_indices", self.sparse_indices)

    def _updated_ctor_param(self):
        """
        Add the two constructor's arguments for the Factory
        """
        dct = super()._updated_ctor_param()
        dct["sparse_indices"] = self.sparse_indices
        dct["xvals"] = self._xvals
        dct["mu"] = self.mu
        dct["sig"] = self.sig
        dct["dims"] = self.dims
        return dct

    @classmethod
    def get_allocation_kwds(
        cls, npdf, **kwargs
    ) -> dict[str, tuple[tuple[int, int], str]]:
        if "dims" not in kwargs:
            raise ValueError("required argument dims not in kwargs")  # pragma: no cover
        nsp = np.array(kwargs["dims"]).flatten()[4]
        return dict(sparse_indices=((npdf, nsp), "i8"))

    @classmethod
    def add_mappings(cls):
        """
        Add this classes mappings to the conversion dictionary
        """
        cls._add_creation_method(cls.create, None)
        cls._add_extraction_method(extract_sparse_from_xy, None)

    @classmethod
    def create_ensemble(
        self, sparse_indices, xvals, mu, sig, dims, ancil: Optional[Mapping] = None
    ) -> Ensemble:
        """Creates an Ensemble of distributions parameterized as interpolations, constructed from a sparse representation.


        Parameters
        ----------
        sparse_indices:
        xvals :
        mu :
        sig :
        dims :
        ancil : Optional[Mapping], optional
            A dictionary of metadata for the distributions, where any arrays have the same length as the number of distributions, by default None

        Returns
        -------
        Ensemble
            An Ensemble object containing all of the given distributions.
        """
        data = {
            "sparse_indices": sparse_indices,
            "xvals": xvals,
            "mu": mu,
            "sig": sig,
            "dims": dims,
        }
        return Ensemble(self, data, ancil)


sparse = sparse_gen

add_class(sparse_gen)



"""This creates data for tests"""

import os
import sys
import numpy as np
import scipy.stats as sps
import pandas as pd
from scipy import interpolate as sciinterp
from scipy import integrate as sciint
from pathlib import Path


import qp
from qp.parameterizations.packed_interp.packing_utils import PackingType, pack_array
from qp.parameterizations.sparse_interp import sparse_rep


###############################################################################
## Variables
###############################################################################

np.random.seed(1234)

NPDF = 11
NBIN = 61
NSAMPLES = 100
XMIN = 0.0
XMAX = 5.0
LOC = np.expand_dims(np.linspace(0.5, 2.5, NPDF), -1)
SCALE = np.expand_dims(np.linspace(0.2, 1.2, NPDF), -1)
LOC_SHIFTED = LOC + SCALE
TEST_XVALS = np.linspace(XMIN, XMAX, 201)
XBINS = np.linspace(XMIN, XMAX, NBIN)
XARRAY = np.ones((NPDF, NBIN)) * XBINS
YARRAY = np.expand_dims(np.linspace(0.5, 2.5, NPDF), -1) * (
    1.0 + 0.1 * np.random.uniform(size=(NPDF, NBIN))
)
HIST_DATA = YARRAY[:, 0:-1]
QUANTS = np.linspace(0.01, 0.99, NBIN)
QLOCS = sps.norm(loc=LOC, scale=SCALE).ppf(QUANTS)
SAMPLES = sps.norm(loc=LOC, scale=SCALE).rvs(size=(NPDF, NSAMPLES))

MEAN_MIXMOD = np.vstack(
    [
        np.linspace(0.5, 2.5, NPDF),
        np.linspace(0.5, 1.5, NPDF),
        np.linspace(1.5, 2.5, NPDF),
    ]
).T
STD_MIXMOD = np.vstack(
    [
        np.linspace(0.2, 1.2, NPDF),
        np.linspace(0.2, 0.5, NPDF),
        np.linspace(0.2, 0.5, NPDF),
    ]
).T
WEIGHT_MIXMOD = np.vstack(
    [0.7 * np.ones((NPDF)), 0.2 * np.ones((NPDF)), 0.1 * np.ones((NPDF))]
).T

HIST_TOL = 4.0 / NBIN
QP_TOPDIR = os.path.dirname(os.path.dirname(__file__))

###############################################################################
## Data construction helper functions
###############################################################################


def calc_ypacked():
    """Writes out test data and parameters for the packed interp parameterization"""

    ypacked_lin, ymax_lin = pack_array(PackingType.linear_from_rowmax, YARRAY.copy())
    ypacked_log, ymax_log = pack_array(
        PackingType.log_from_rowmax, YARRAY.copy(), log_floor=-3
    )

    return ypacked_lin, ymax_lin, ypacked_log, ymax_log


def get_sparse_data():
    """Writes out test data and parameters for the sparse parameterization"""

    TEST_DIR = Path(__file__).resolve().parent
    filein = os.path.join(TEST_DIR.parent, "test_data", "CFHTLens_sample.P.npy")

    # FORMAT FILE, EACH ROW IS THE PDF FOR EACH GALAXY, LAST ROW IS THE REDSHIFT POSITION
    P = np.load(filein)
    z = P[-1]
    P = P[:NPDF]
    P = P / sciint.trapezoid(P, z).reshape(-1, 1)
    minz = np.min(z)
    nz = 301
    _, j = np.where(P > 0)
    maxz = np.max(z[j + 1])
    newz = np.linspace(minz, maxz, nz)
    interp = sciinterp.interp1d(z, P, assume_sorted=True)
    newpdf = interp(newz)
    newpdf = newpdf / sciint.trapezoid(newpdf, newz).reshape(-1, 1)
    sparse_idx, meta, _ = sparse_rep.build_sparse_representation(
        newz, newpdf, verbose=False
    )

    return sparse_idx, meta


###############################################################################
## The test data dictionaries
###############################################################################


hist_test_data = dict(
    hist=dict(
        gen_func=qp.hist,
        ctor_data=dict(bins=XBINS, pdfs=HIST_DATA),
        convert_data=dict(bins=XBINS),
        atol_diff=1e-1,
        atol_diff2=1e-1,
        test_xvals=TEST_XVALS,
    ),
    hist_samples=dict(
        gen_func=qp.hist,
        filekey="hist_samples",
        ctor_data=dict(bins=XBINS, pdfs=HIST_DATA),
        convert_data=dict(
            bins=XBINS,
            method="samples",
            size=NSAMPLES,
        ),
        atol_diff=1e-1,
        atol_diff2=1e-1,
        test_xvals=TEST_XVALS,
        do_samples=True,
    ),
)


interp_test_data = dict(
    interp=dict(
        gen_func=qp.interp,
        ctor_data=dict(xvals=XBINS, yvals=YARRAY),
        convert_data=dict(xvals=XBINS),
        test_xvals=TEST_XVALS,
    )
)
interp_irregular_test_data = dict(
    interp_irregular=dict(
        gen_func=qp.interp_irregular,
        ctor_data=dict(xvals=XARRAY, yvals=YARRAY),
        convert_data=dict(xvals=XBINS),
        test_xvals=TEST_XVALS,
    )
)

mixmod_test_data = dict(
    mixmod=dict(
        gen_func=qp.mixmod,
        ctor_data=dict(
            weights=WEIGHT_MIXMOD,
            means=MEAN_MIXMOD,
            stds=STD_MIXMOD,
        ),
        convert_data={},
        test_xvals=TEST_XVALS,
        atol_diff2=1.0,
    )
)
quant_test_data = dict(
    quant=dict(
        gen_func=qp.quant,
        ctor_data=dict(quants=QUANTS, locs=QLOCS),
        convert_data=dict(quants=QUANTS),
        test_xvals=TEST_XVALS,
    )
)

norm_test_data = dict(
    norm=dict(
        gen_func=qp.stats.norm,
        ctor_data=dict(loc=LOC, scale=SCALE),
        test_xvals=TEST_XVALS,
        do_samples=True,
        ancil=dict(zmode=LOC),
    ),
    norm_shifted=dict(
        gen_func=qp.stats.norm,
        filekey="norm_shifted",
        ctor_data=dict(loc=LOC, scale=SCALE),
        test_xvals=TEST_XVALS,
    ),
    norm_multi_d=dict(
        gen_func=qp.stats.norm,
        filekey="norm_multi_d",
        ctor_data=dict(
            loc=np.array([LOC, LOC]),
            scale=np.array([SCALE, SCALE]),
        ),
        test_xvals=TEST_XVALS,
        do_samples=True,
        npdf=22,
    ),
)

ypacked_lin, ymax_lin, ypacked_log, ymax_log = calc_ypacked()

packed_interp_test_data = dict(
    lin_packed_interp=dict(
        gen_func=qp.packed_interp,
        ctor_data=dict(
            packing_type=PackingType.linear_from_rowmax,
            xvals=XBINS,
            ypacked=ypacked_lin,
            ymax=ymax_lin,
        ),
        convert_data=dict(
            xvals=XBINS,
            packing_type=PackingType.linear_from_rowmax,
        ),
        test_xvals=TEST_XVALS,
    ),
    log_packed_interp=dict(
        gen_func=qp.packed_interp,
        ctor_data=dict(
            packing_type=PackingType.log_from_rowmax,
            xvals=XBINS,
            ypacked=ypacked_log,
            ymax=ymax_log,
            log_floor=-3.0,
        ),
        convert_data=dict(
            xvals=XBINS,
            packing_type=PackingType.log_from_rowmax,
            log_floor=-3.0,
        ),
        test_xvals=TEST_XVALS,
    ),
)

sparse_idx, meta = get_sparse_data()

sparse_test_data = dict(
    sparse=dict(
        gen_func=qp.sparse,
        ctor_data=dict(
            xvals=meta["xvals"],
            mu=meta["mu"],
            sig=meta["sig"],
            dims=meta["dims"],
            sparse_indices=sparse_idx,
        ),
        test_xvals=TEST_XVALS,
    ),
)

SPLX, SPLY, SPLN = qp.spline.build_normed_splines(XARRAY, YARRAY)
spline_test_data = dict(
    spline=dict(
        gen_func=qp.spline,
        ctor_data=dict(splx=SPLX, sply=SPLY, spln=SPLN),
        test_xvals=TEST_XVALS[::10],
    ),
    spline_kde=dict(
        gen_func=qp.spline,
        method="samples",
        ctor_data=dict(samples=SAMPLES, xvals=np.linspace(0, 5, 51)),
        convert_data=dict(xvals=np.linspace(0, 5, 51), method="samples"),
        test_xvals=TEST_XVALS,
        atol_diff2=1.0,
        test_pdf=False,
    ),
    spline_xy=dict(
        gen_func=qp.spline,
        method="xy",
        ctor_data=dict(xvals=XARRAY, yvals=YARRAY),
        convert_data=dict(xvals=np.linspace(0, 5, 51), method="xy"),
        test_xvals=TEST_XVALS,
        test_pdf=False,
    ),
)



"""
Unit tests for PDF class
"""

import os
import unittest

import h5py
import numpy as np
import pytest
from mpi4py import MPI
import tempfile
from pathlib import Path

import qp
from tests.helpers.test_funcs import build_ensemble
from tests.helpers import test_data_helper as t_data


@pytest.mark.skipif(
    not h5py.get_config().mpi, reason="Do not have parallel version of hdf5py"
)
class EnsembleTestCase(unittest.TestCase):
    """Class to test qp.Ensemble functionality"""

    def setUp(self):
        """
        Make any objects that are used in multiple tests.
        """
        self.tmpdir = tempfile.TemporaryDirectory()

        self.norm_test_data = dict(
            norm=dict(
                gen_func=qp.stats.norm,
                ctor_data=dict(loc=t_data.LOC, scale=t_data.SCALE),
                test_xvals=t_data.TEST_XVALS,
                do_samples=True,
                ancil=dict(zmode=t_data.LOC),
            ),
            norm_shifted=dict(
                gen_func=qp.stats.norm,
                filekey="norm_shifted",
                ctor_data=dict(loc=t_data.LOC, scale=t_data.SCALE),
                test_xvals=t_data.TEST_XVALS,
            ),
            norm_multi_d=dict(
                gen_func=qp.stats.norm,
                filekey="norm_multi_d",
                ctor_data=dict(
                    loc=np.array([t_data.LOC, t_data.LOC]),
                    scale=np.array([t_data.SCALE, t_data.SCALE]),
                ),
                test_xvals=t_data.TEST_XVALS,
                do_samples=True,
                npdf=22,
            ),
        )
        # make hist test data
        self.hist_test_data = dict(
            hist=dict(
                gen_func=qp.hist,
                ctor_data=dict(bins=t_data.XBINS, pdfs=t_data.HIST_DATA),
                convert_data=dict(bins=t_data.XBINS),
                atol_diff=1e-1,
                atol_diff2=1e-1,
                test_xvals=t_data.TEST_XVALS,
            ),
            hist_samples=dict(
                gen_func=qp.hist,
                filekey="hist_samples",
                ctor_data=dict(bins=t_data.XBINS, pdfs=t_data.HIST_DATA),
                convert_data=dict(
                    bins=t_data.XBINS, method="samples", size=t_data.NSAMPLES
                ),
                atol_diff=1e-1,
                atol_diff2=1e-1,
                test_xvals=t_data.TEST_XVALS,
                do_samples=True,
            ),
        )
        self.interp_test_data = dict(
            interp=dict(
                gen_func=qp.interp,
                ctor_data=dict(xvals=t_data.XBINS, yvals=t_data.YARRAY),
                convert_data=dict(xvals=t_data.XBINS),
                test_xvals=t_data.TEST_XVALS,
            )
        )

    def tearDown(self):
        "Clean up any mock data files created by the tests."

        self.tmpdir.cleanup()

    @staticmethod
    def _run_ensemble_funcs(ens_type, ens, _xpts, tmpdir):
        """Run the test for a practicular class"""
        comm = MPI.COMM_WORLD  # pylint: disable=c-extension-no-member
        mpi_rank = comm.Get_rank()
        mpi_size = comm.Get_size()
        zmode = ens.mode(grid=np.linspace(-3, 3, 100))
        ens.set_ancil(dict(zmode=zmode, ones=np.ones(ens.npdf)))

        filename = Path(tmpdir) / f"testwrite_{ens_type}.hdf5"
        group, fout = ens.initializeHdf5Write(filename, ens.npdf * mpi_size, comm)
        ens.writeHdf5Chunk(group, mpi_rank * ens.npdf, (mpi_rank + 1) * ens.npdf)
        ens.finalizeHdf5Write(fout)

        readens = qp.read(filename)
        assert sum(readens.ancil["ones"]) == mpi_size * ens.npdf
        assert len(readens.ancil["zmode"]) == mpi_size * ens.npdf
        assert readens.metadata.keys() == ens.metadata.keys()
        assert readens.objdata.keys() == ens.objdata.keys()

        test_grid = np.linspace(-3, 3, 100)
        itr = qp.iterator(filename, 10, mpi_rank, mpi_size)
        for start, end, ens_i in itr:
            assert np.allclose(readens[start:end].pdf(test_grid), ens_i.pdf(test_grid))

        # if mpi_rank == 0:
        #     os.remove(f"testwrite_{ens_type}.hdf5")

    def test_parallel_norm(self):
        """Run the ensemble tests on an ensemble of scipy.stats.norm distributions"""
        key = "norm"
        cls_test_data = self.norm_test_data[key]  # pylint: disable=no-member
        ens_norm = build_ensemble(cls_test_data)
        self._run_ensemble_funcs(
            "norm", ens_norm, cls_test_data["test_xvals"], self.tmpdir.name
        )

    def test_parallel_hist(self):
        """Run the ensemble tests on an ensemble of qp.hist distributions"""
        key = "hist"
        # qp.hist_gen.make_test_data()
        cls_test_data = self.hist_test_data[key]
        ens_h = build_ensemble(cls_test_data)
        self._run_ensemble_funcs(
            "hist", ens_h, cls_test_data["test_xvals"], self.tmpdir.name
        )

    def test_parallel_interp(self):
        """Run the ensemble tests on an ensemble of qp.interp distributions"""
        key = "interp"
        # qp.interp_gen.make_test_data()
        cls_test_data = self.interp_test_data[key]
        ens_i = build_ensemble(cls_test_data)
        self._run_ensemble_funcs(
            "interp", ens_i, cls_test_data["test_xvals"], self.tmpdir.name
        )


if __name__ == "__main__":
    unittest.main()



"""This module implements a distribution parameterization sub-class using histograms"""

from __future__ import annotations
import numpy as np

from scipy.stats import rv_continuous
from typing import Mapping, Optional
from numpy.typing import ArrayLike


import warnings

from .hist_utils import (
    evaluate_hist_x_multi_y,
    extract_hist_values,
    extract_hist_samples,
)
from ..base import Pdf_rows_gen
from ...plotting import get_axes_and_xlims, plot_pdf_histogram_on_axes
from ...utils.array import reshape_to_pdf_size, reduce_dimensions

from ...utils.interpolation import interpolate_multi_x_y, interpolate_x_multi_y

from ...core.factory import add_class
from ...core.ensemble import Ensemble


class hist_gen(Pdf_rows_gen):
    """Implements distributions parameterized as histograms.

    By default, the input distribution is normalized. If the input data is
    already normalized, you can use the optional parameter ``norm = False``
    to skip the normalization process.

    Parameters
    ----------
    bins : ArrayLike
        The array containing the (n+1) bin boundaries
    pdfs : ArrayLike
        The array containing the (npdf, n) bin values
    norm : bool, optional
        If True, normalizes the input distribution. If False, assumes the
        given distribution is already normalized. By default True.
    warn : bool, optional
            If True, raises warnings if input is not valid PDF data (i.e. if
            data is negative). If False, no warnings are raised. By default True.


    Notes
    -----

    There must be a minimum of 2 bins.

    Converting to this parameterization:

    This table contains the available methods to convert to this parameterization,
    their required arguments, and their method keys. If the key is `None`, this is
    the default conversion method.

    +------------------------+-----------------------------------------------------+------------+
    | Function               | Arguments                                           | Method key |
    +------------------------+-----------------------------------------------------+------------+
    | `.extract_hist_values` | bins (array of bin edges)                           | None       |
    +------------------------+-----------------------------------------------------+------------+
    | `.extract_hist_samples`| bins (array of bin edges),                          | samples    |
    |                        | size (int, optional, number of samples to generate) |            |
    +------------------------+-----------------------------------------------------+------------+

    Implementation notes:

    Inside a given bin `pdf()` will return the `hist_gen.pdfs` value.
    Outside the range of the given bins `pdf()` will return 0.

    Inside a given bin `cdf()` will use a linear interpolation across the bin.
    Outside the range of the given bins `cdf()` will return (0 or 1), respectively.

    The percentage point function `ppf()` will return negative infinity at 0 and positive
    infinity at 1.

    """

    # pylint: disable=protected-access

    name = "hist"
    version = 0

    _support_mask = rv_continuous._support_mask

    def __init__(
        self,
        bins: ArrayLike,
        pdfs: ArrayLike,
        norm: bool = True,
        warn: bool = True,
        *args,
        **kwargs,
    ) -> None:
        """
        Create a new distribution using the given histogram.

        Parameters
        ----------
        bins : ArrayLike
          The array containing the (n+1) bin boundaries
        pdfs : ArrayLike
          The array containing the (npdf, n) bin values
        norm : bool, optional
            If True, normalizes the input distribution. If False, assumes the
            given distribution is already normalized. By default True.
        warn : bool, optional
            If True, raises warnings if input is not valid PDF data (i.e. if
            data is negative). If False, no warnings are raised. By default True.
        """
        self._hbins = np.squeeze(bins)  # make sure bins is 1D
        self._nbins = self._hbins.size - 1
        self._hpdfs = reshape_to_pdf_size(np.asarray(pdfs), -1)

        # make sure that the bins are sorted
        if not np.all(np.diff(self._hbins) >= 0):
            raise ValueError(
                f"Invalid bins: The given bins are not sorted: {self._hbins}"
            )

        # raise warnings if input data is not finite or pdfs are not positive
        self._warn = warn
        if self._warn:
            if not np.all(np.isfinite(self._hbins)):
                warnings.warn(
                    f"The given bins contain non-finite values - {self._hbins}",
                    RuntimeWarning,
                )
            if not np.all(np.isfinite(pdfs)):
                indices = np.where(np.isfinite(pdfs) != True)
                warnings.warn(
                    f"There are non-finite values in the pdfs for the distributions: {indices[0]}",
                    RuntimeWarning,
                )
            if np.any(self._hpdfs < 0):
                indices = np.where(self._hpdfs < 0)
                warnings.warn(
                    f"There are negative values in the pdfs for the distributions: {indices[0]}",
                    RuntimeWarning,
                )

        # check data shapes make sense
        if np.shape(pdfs)[-1] != self._nbins:  # pragma: no cover
            raise ValueError(
                "Number of bins (%i) != number of values (%i)"
                % (self._nbins, np.shape(pdfs)[-1])
            )

        self._hbin_widths = self._hbins[1:] - self._hbins[:-1]
        self._xmin = self._hbins[0]
        self._xmax = self._hbins[-1]

        # normalize the input data if norm is True
        self._norm = norm
        if self._norm:
            self._hpdfs = self.normalize()["pdfs"]

        self._hcdfs = None
        # Set support
        kwargs["shape"] = self._hpdfs.shape  # pdfs.shape
        super().__init__(*args, **kwargs)
        self._addmetadata("bins", self._hbins)
        self._addobjdata("pdfs", self._hpdfs)

    def _compute_cdfs(self):
        copy_shape = np.array(self._hpdfs.shape)
        copy_shape[-1] += 1
        self._hcdfs = np.ndarray(copy_shape)
        self._hcdfs[:, 0] = 0.0
        self._hcdfs[:, 1:] = np.cumsum(self._hpdfs * self._hbin_widths, axis=1)

    def normalize(self) -> Mapping[str, np.ndarray[float]]:
        """Normalizes the input distribution values.

        Returns
        -------
        Mapping [str, np.ndarray[float]]
            An (npdf, n) array of pdf values in the n bins for the npdf distributions

        Raises
        ------
        ValueError
            Raised if the sum under the distribution <= 0.
        """
        pdfs_2d = self._hpdfs
        sums = np.sum(pdfs_2d * self._hbin_widths, axis=1)
        if np.any(sums < 0):
            indices = np.where(sums < 0)
            raise ValueError(
                f"The distribution(s) cannot be properly normalized, the sum of the pdfs is < 0 for distributions at index = {indices[0]} "
            )
        elif np.any(sums == 0):
            indices = np.where(sums == 0)
            warnings.warn(
                f"The distributions(s) with indices {indices[0]} have an integral of 0."
            )
        return {"pdfs": (pdfs_2d.T / sums).T}

    @property
    def bins(self) -> np.ndarray[float]:
        """Return the histogram bin edges"""
        return self._hbins

    @property
    def pdfs(self) -> np.ndarray[float]:
        """Return the histogram bin values"""
        return self._hpdfs

    def x_samples(self) -> np.ndarray[float]:
        """Return a set of x values that can be used to plot all the PDFs."""
        # TODO: possibly add a bin to the left and right?
        return (self._hbins[:-1] + self._hbins[1:]) / 2

    def _pdf(self, x, row):
        # pylint: disable=arguments-differ
        pdf = evaluate_hist_x_multi_y(x, row, self._hbins, self._hpdfs).ravel()

        # reduce dimension to 0 if there's only one value
        if np.shape(pdf) == (1,) and len(pdf) == 1:
            return pdf[0]
        else:
            return pdf

    def _cdf(self, x, row):
        # pylint: disable=arguments-differ
        if self._hcdfs is None:  # pragma: no cover
            self._compute_cdfs()
        return interpolate_x_multi_y(
            x, row, self._hbins, self._hcdfs, bounds_error=False, fill_value=(0.0, 1.0)
        ).ravel()

    def _ppf(self, x, row):
        # pylint: disable=arguments-differ
        if self._hcdfs is None:  # pragma: no cover
            self._compute_cdfs()
        return interpolate_multi_x_y(
            x,
            row,
            self._hcdfs,
            self._hbins,
            bounds_error=False,
            fill_value=(self._xmin, self._xmax),
        ).ravel()

    def _munp(self, m, *args):
        """compute moments"""
        # pylint: disable=arguments-differ
        # Silence floating point warnings from integration.
        with np.errstate(all="ignore"):
            vals = self.custom_generic_moment(m)
        return vals

    def custom_generic_moment(self, m: ArrayLike) -> np.ndarray[float]:
        """Compute the mth moment"""
        m = np.asarray(m)
        dx = self._hbins[1] - self._hbins[0]
        xv = 0.5 * (self._hbins[1:] + self._hbins[:-1])
        return np.sum(xv**m * self._hpdfs, axis=1) * dx

    def _updated_ctor_param(self):
        """
        Set the bins as additional constructor argument
        """
        dct = super()._updated_ctor_param()
        dct["bins"] = self._hbins
        dct["pdfs"] = self._hpdfs
        dct["norm"] = self._norm
        dct["warn"] = self._warn
        return dct

    @classmethod
    def get_allocation_kwds(
        cls, npdf: int, **kwargs: str
    ) -> dict[str, tuple[tuple[int, int], str]]:
        """Return the kwds necessary to create an `empty` HDF5 file with ``npdf`` entries
        for iterative write. We only need to allocate the data columns, as
        the metadata will be written when we finalize the file.

        The number of data columns is calculated based on the length or shape of the
        metadata, ``n``. For example, the number of columns is ``nbins-1``
        for a histogram.

        Parameters
        ----------
        npdf : int
            Total number of distributions that will be written out
        kwargs :
            The keys needed to construct the shape of the data to be written.

        Returns
        -------
        dict [ str, tuple [ tuple [int , int], str]]
            A dictionary with a key for the objdata, a tuple with the shape of that data,
            and the data type of the data as a string.
            i.e. ``{objdata_key = ( (npdf, n), "f4" )}``

        Raises
        ------
        ValueError
            Raises an error if the bins is not provided."""

        if "bins" not in kwargs:  # pragma: no cover
            raise ValueError("required argument 'bins' not included in kwargs")
        nbins = len(kwargs["bins"].flatten())
        return dict(pdfs=((npdf, nbins - 1), "f4"))

    @classmethod
    def plot_native(cls, pdf: Ensemble, **kwargs):
        """Plot the PDF in a way that is particular to this type of distribution

        For a histogram this shows the bin edges.

        Parameters
        ----------
        axes : Axes
            The axes to plot on. Either this or xlim must be provided.
        xlim : tuple [float, float]
            The x-axis limits. Either this or axes must be provided.
        kwargs :
            Any keyword arguments to pass to matplotlib's `matplotlib.axes.Axes.hist` method.

        Returns
        -------
        axes : Axes
            The plot axes.
        """
        axes, _, kw = get_axes_and_xlims(**kwargs)
        vals = pdf.dist.pdfs[pdf.kwds["row"]]
        return plot_pdf_histogram_on_axes(axes, hist=(pdf.dist.bins, vals), **kw)

    @classmethod
    def add_mappings(cls):
        """
        Add this classes mappings to the conversion dictionary
        """
        cls._add_creation_method(cls.create, None)
        cls._add_extraction_method(extract_hist_values, None)
        cls._add_extraction_method(extract_hist_samples, "samples")

    @classmethod
    def create_ensemble(
        self,
        bins: ArrayLike,
        pdfs: ArrayLike,
        norm: bool = True,
        warn: bool = True,
        ancil: Optional[Mapping] = None,
    ) -> Ensemble:
        """Creates an Ensemble of distributions parameterized as histograms.


        Parameters
        ----------
        bins : ArrayLike
          The array containing the (n+1) bin boundaries
        pdfs : ArrayLike
          The array containing the (npdf, n) bin values
        norm : bool, optional
            If True, normalizes the input distribution. If False, assumes the
            given distribution is already normalized. By default True.
        warn : bool, optional
            If True, raises warnings if input is not valid PDF data (i.e. if
            data is negative). If False, no warnings are raised. By default True.
        ancil : Optional[Mapping], optional
            A dictionary of metadata for the distributions, where any arrays have
            length npdf, by default None

        Returns
        -------
        Ensemble
            An Ensemble object containing all of the given distributions.

        Examples
        --------

        To create an Ensemble with two distributions and an 'ancil' table that
        provides ids for the distributions, you can use the following code:

        >>> import qp
        >>> import numpy as np
        >>> bins= [0,1,2,3,4,5]
        >>> pdfs = np.array([[0,0.1,0.1,0.4,0.2],[0.05,0.09,0.2,0.3,0.15]])
        >>> ancil = {'ids': [105, 108]}
        >>> ens = qp.hist.create_ensemble(bins,pdfs,ancil=ancil)
        >>> ens.metadata
        {'pdf_name': array([b'hist'], dtype='|S4'),
        'pdf_version': array([0]),
        'bins': array([[0, 1, 2, 3, 4, 5]])}

        """

        data = {"bins": bins, "pdfs": pdfs, "norm": norm, "warn": warn}
        return Ensemble(self, data, ancil)


hist = hist_gen
add_class(hist_gen)



"""Module docstring.

Note that this module has something to do with `myPackage.otherModule.MyOtherClass`
which we can format and link with single backticks.

Alternatively, to only show the final pathe element, a ~ can be added:
`~myPackage.otherModule.MyOtherClass`.

We can link to any external Python function or class (from a package included in the
intersphinx configuration) by adding the full path: `scipy.stats.norm`. The package does
not need to be imported.

In any case, to provide a link title: `title <MyClass>`.
"""

#### Example Module


class MyClass:
    """Class docstring.

    This will appear in the RDT documentation.

    Parameters
    ----------
    parameter : type
        description

    Examples
    --------
    >>> from module import MyClass
    >>> print("Examples should be plural!")
    Examples should be plural!

    """

    def __init__(self, parameter: type) -> None:
        """Constructor docstring. This will not appear in the RTD documentation.

        Parameters
        ----------
        parameter : type
            description

        """
        pass

    def method(self: "MyClass", dictionary: dict[str, int | float]) -> str:
        """Instance method.

        Parameters
        ----------
        self : MyClass
            No formatting needs to be done on "MyClass" since this is the module where
            it is defined.
        dictionary : dict[str , int | float]
            Note the spaces inside the brackets around the comma and pipe.

        Returns
        -------
        str
            Instead of the type/description pattern used here, the Returns section also
            accepts the name/type/description pattern used in the parameters section.

        """



import pytest
import qp
import numpy as np


@pytest.fixture
def mixmod_ensemble(mixmod_test_data) -> qp.Ensemble:
    ens = qp.mixmod.create_ensemble(**mixmod_test_data["mixmod"]["ctor_data"])
    return ens


def test_norm(mixmod_ensemble):
    """Test that calling norm raises the appropriate error."""

    with pytest.raises(
        RuntimeError,
        match="The distributions in a mixmod parameterization are already normalized",
    ):
        mixmod_ensemble.norm()


def test_x_samples(mixmod_ensemble):
    """Test that x_samples works as expected."""

    xsamps = mixmod_ensemble.x_samples()

    xmin = np.min(mixmod_ensemble.objdata["means"]) - np.max(
        mixmod_ensemble.objdata["stds"]
    )
    xmax = np.max(mixmod_ensemble.objdata["means"]) + np.max(
        mixmod_ensemble.objdata["stds"]
    )

    assert np.min(xsamps) <= xmin
    assert np.max(xsamps) >= xmax
    assert len(xsamps) >= 50
    assert len(xsamps) <= 10000


@pytest.mark.parametrize(
    ("means_big,stds_small, stds_big"),
    [((-25, 25), (3, 7), (25, 35)), ((-100, 100), (0.05, 0.5), (50, 100))],
)
def test_x_samples_with_more_npts(means_big, stds_small, stds_big):
    """Make sure that the x_smaples method works when npts is used and when npts_max is used."""

    npdf = 5

    mean = np.vstack(
        [
            np.linspace(-2, 2, npdf),
            np.linspace(-10, 10, npdf),
            np.linspace(*means_big, num=npdf),
        ]
    ).T
    std = np.vstack(
        [
            np.linspace(*stds_small, num=npdf),
            np.linspace(5, 10, npdf),
            np.linspace(*stds_big, num=npdf),
        ]
    ).T
    weights = np.vstack(
        [0.7 * np.ones((npdf)), 0.2 * np.ones((npdf)), 0.1 * np.ones((npdf))]
    ).T

    ens_m = qp.mixmod.create_ensemble(means=mean, stds=std, weights=weights)

    xmin = np.min(ens_m.objdata["means"]) - np.max(ens_m.objdata["stds"])
    xmax = np.max(ens_m.objdata["means"]) + np.max(ens_m.objdata["stds"])

    dx = np.min(ens_m.objdata["stds"]) / 2.0
    npts = (xmax - xmin) // dx

    xsamps = ens_m.x_samples()

    # make sure limits of xsamples makes sense
    assert xmin == np.min(xsamps)
    assert xmax == np.max(xsamps)

    # make sure the number of points makes sense
    if npts > 10000:
        assert len(xsamps) == 10000
    else:
        assert len(xsamps) == npts


@pytest.mark.parametrize(
    "means, stds, weights, match_string",
    [
        (
            np.array([0, 0.5, np.nan]),
            np.array([0.25, 0.5, 1.0]),
            np.array([1, 1, 1]),
            "The given means contain non-finite values for the following distributions",
        ),
        (
            np.array([0, 0.5, 1]),
            np.array([0.25, 0.5, np.inf]),
            np.array([1, 0.5, 1]),
            "here are non-finite values in the stds for the following distributions",
        ),
        (
            np.array([0, 0.5, 1]),
            np.array([0.25, 0.5, 1.0]),
            np.array([1, 1, np.nan]),
            "There are non-finite values in the weights for the following distributions",
        ),
        (
            np.array([0, 0.5, 1]),
            np.array([0.0, 0.0, 0.0]),
            np.array([0.0, 1.0, 0.0]),
            "The following distributions have all stds",
        ),
    ],
)
def test_warnings(means, stds, weights, match_string):
    """Test that warnings when building a mixmod Ensemble work."""

    with pytest.warns(RuntimeWarning, match=match_string):
        ens = qp.mixmod.create_ensemble(means=means, stds=stds, weights=weights)


@pytest.mark.parametrize(
    "means, stds, weights, match_string",
    [
        (
            np.array([0, 0.5]),
            np.array([0.25, 0.5, 1.0]),
            np.array([1, 1, 1]),
            "must have the same shape",
        ),
        (
            np.array([0, 0.5, 1.0]),
            np.array([0.25, 0.5, 1.0]),
            np.array([-0.2, 0, 0.5]),
            "Invalid input: All weights need to be larger than or equal to 0",
        ),
        (
            np.array([0, 0.5, 1.0]),
            np.array([0.25, -0.5, 1.0]),
            np.array([1, 0.5, 0.5]),
            "Invalid input: All standard deviations",
        ),
    ],
)
def test_invalid_input(means, stds, weights, match_string):
    """Tests that the appropriate errors are raised when invalid input is given to build a mixmod Ensemble."""

    with pytest.raises(ValueError, match=match_string):
        ens = qp.mixmod.create_ensemble(means=means, stds=stds, weights=weights)



from __future__ import annotations

import numpy as np
from numpy.typing import ArrayLike
from scipy.interpolate import interp1d

from .array import get_eval_case, CASE_PRODUCT, CASE_FACTOR, CASE_2D, CASE_FLAT


def interpolate_multi_x_y(
    x: ArrayLike, row: ArrayLike, xvals: ArrayLike, yvals: ArrayLike, **kwargs
) -> np.ndarray:
    """
    Interpolate a set of values

    Parameters
    ----------
    x : ArrayLike, shape (npdf, n)
        X values to interpolate at
    row : ArrayLike, shape (npdf, 1)
        Which rows to interpolate at
    xvals : ArrayLike, shape (npdf, npts)
        X-values used for the interpolation
    yvals : ArrayLike, shape (npdf)
        Y-values used for the interpolation

    Returns
    -------
    vals : np.ndarray
        The interpolated values
    """
    case_idx, xx, rr = get_eval_case(x, row)
    if case_idx in [CASE_PRODUCT, CASE_FACTOR]:
        return interpolate_multi_x_y_product(xx, rr, xvals, yvals, **kwargs)
    if case_idx == CASE_2D:
        return interpolate_multi_x_y_2d(xx, rr, xvals, yvals, **kwargs)
    return interpolate_multi_x_y_flat(xx, rr, xvals, yvals, **kwargs)


def interpolate_multi_x_y_product(
    x: ArrayLike, row: ArrayLike, xvals: ArrayLike, yvals: ArrayLike, **kwargs
) -> np.ndarray:
    """
    Interpolate a set of values

    Parameters
    ----------
    x : ArrayLike, length n
        X values to interpolate at
    row : ArrayLike, shape (npdf, 1)
        Which rows to interpolate at
    xvals : ArrayLike, shape (npdf, npts)
        X-values used for the interpolation
    yvals : ArrayLike, length npdf
        Y-values used for the interpolation

    Returns
    -------
    vals : np.ndarray, shape (npdf, n)
        The interpolated values
    """
    rr = np.squeeze(row)
    nx = np.shape(x)[-1]

    def single_row(rv):
        return interp1d(xvals[rv], yvals, **kwargs)(x)

    vv = np.vectorize(single_row, signature="()->(%i)" % (nx))
    return vv(rr)


def interpolate_multi_x_y_2d(
    x: ArrayLike, row: ArrayLike, xvals: ArrayLike, yvals: ArrayLike, **kwargs
) -> np.ndarray:
    """
    Interpolate a set of values

    Parameters
    ----------
    x : ArrayLike, shape (npdf, n)
        X values to interpolate at
    row : ArrayLike, shape (npdf, 1)
        Which rows to interpolate at
    xvals : ArrayLike, shape (npdf, npts)
        X-values used for the interpolation
    yvals : ArrayLike, length npdf
        Y-values used for the interpolation

    Returns
    -------
    vals : np.ndarray, shape (npdf, n)
        The interpolated values
    """
    nx = np.shape(x)[-1]

    def evaluate_row(rv, xv):
        return interp1d(xvals[rv], yvals, **kwargs)(xv)

    vv = np.vectorize(evaluate_row, signature="(),(%i)->(%i)" % (nx, nx))
    return vv(np.squeeze(row), x)


def interpolate_multi_x_y_flat(
    x: ArrayLike, row: ArrayLike, xvals: ArrayLike, yvals: ArrayLike, **kwargs
) -> np.ndarray:
    """
    Interpolate a set of values

    Parameters
    ----------
    x : ArrayLike, length n
        X values to interpolate at
    row : ArrayLike, length n
        Which rows to interpolate at
    xvals : ArrayLike, shape (npdf, npts)
        X-values used for the interpolation
    yvals : ArrayLike, length npdf
        Y-values used for the interpolation

    Returns
    -------
    vals : np.ndarray, shape (npdf, n)
        The interpolated values
    """

    def single_row(xv, rv):
        return interp1d(xvals[rv], yvals, **kwargs)(xv)

    vv = np.vectorize(single_row)
    return vv(x, row)


def interpolate_x_multi_y_product(
    x: ArrayLike, row: ArrayLike, xvals: ArrayLike, yvals: ArrayLike, **kwargs
) -> np.ndarray:
    """
    Interpolate a set of values

    Parameters
    ----------
    x : ArrayLike, length n
        X values to interpolate at
    row : ArrayLike, shape (npdf, 1)
        Which rows to interpolate at
    xvals : ArrayLike, length npts
        X-values used for the interpolation
    yvals : ArrayLike, shape (npdf, npts)
        Y-values used for the interpolation

    Returns
    -------
    vals : np.ndarray, shape (npdf, n)
        The interpolated values
    """
    rr = np.squeeze(row)
    return interp1d(xvals, yvals[rr], **kwargs)(x)


def interpolate_x_multi_y(
    x: ArrayLike, row: ArrayLike, xvals: ArrayLike, yvals: ArrayLike, **kwargs
) -> np.ndarray:
    """
    Interpolate a set of values

    Parameters
    ----------
    x : ArrayLike, shape (npdf, n)
        X values to interpolate at
    row : ArrayLike, shape (npdf, 1)
        Which rows to interpolate at
    xvals : ArrayLike, length npts
        X-values used for the interpolation
    yvals : ArrayLike, shape (npdf, npts)
        Y-values used for the interpolation

    Returns
    -------
    vals : np.ndarray
        The interpolated values
    """
    case_idx, xx, rr = get_eval_case(x, row)
    if case_idx in [CASE_PRODUCT, CASE_FACTOR]:
        return interpolate_x_multi_y_product(xx, rr, xvals, yvals, **kwargs)
    if case_idx == CASE_2D:
        return interpolate_x_multi_y_2d(xx, rr, xvals, yvals, **kwargs)
    return interpolate_x_multi_y_flat(xx, rr, xvals, yvals, **kwargs)


def interpolate_x_multi_y_2d(
    x: ArrayLike, row: ArrayLike, xvals: ArrayLike, yvals: ArrayLike, **kwargs
) -> np.ndarray:
    """
    Interpolate a set of values

    Parameters
    ----------
    x : ArrayLike, shape (npdf, n)
        X values to interpolate at
    row : ArrayLike, shape (npdf, 1)
        Which rows to interpolate at
    xvals : ArrayLike, length npts
        X-values used for the interpolation
    yvals : ArrayLike, shape (npdf, npts)
        Y-values used for the interpolation

    Returns
    -------
    vals : np.ndarray, shape (npdf, n)
        The interpolated values
    """
    nx = np.shape(x)[-1]

    def evaluate_row(rv, xv):
        return interp1d(xvals, yvals[rv], **kwargs)(xv)

    vv = np.vectorize(evaluate_row, signature="(1),(%i)->(%i)" % (nx, nx))
    return vv(row, x)


def interpolate_x_multi_y_flat(
    x: ArrayLike, row: ArrayLike, xvals: ArrayLike, yvals: ArrayLike, **kwargs
) -> np.ndarray:
    """
    Interpolate a set of values

    Parameters
    ----------
    x : ArrayLike, length n
        X values to interpolate at
    row : ArrayLike, length n
        Which rows to interpolate at
    xvals : ArrayLike, length npts
        X-values used for the interpolation
    yvals : ArrayLike, shape (npdf, npts)
        Y-values used for the interpolation

    Returns
    -------
    vals : np.ndarray, shape (npdf, n)
        The interpolated values
    """

    def single_row(xv, rv):
        return interp1d(xvals, yvals[rv], **kwargs)(xv)

    vv = np.vectorize(single_row)
    return vv(x, row)


def interpolate_multi_x_multi_y_flat(
    x: ArrayLike, row: ArrayLike, xvals: ArrayLike, yvals: ArrayLike, **kwargs
) -> np.ndarray:
    """
    Interpolate a set of values

    Parameters
    ----------
    x : ArrayLike, length n
        X values to interpolate at
    row : ArrayLike, length n
        Which rows to interpolate at
    xvals : ArrayLike, shape (npdf, npts)
        X-values used for the interpolation
    yvals : ArrayLike, shape (npdf, npts)
        Y-values used for the interpolation

    Returns
    -------
    vals : np.ndarray, shape (npdf, n)
        The interpolated values
    """

    def single_row(xv, rv):
        return interp1d(xvals[rv], yvals[rv], **kwargs)(xv)

    vv = np.vectorize(single_row)
    return vv(x, row)


def interpolate_multi_x_multi_y_product(
    x: ArrayLike, row: ArrayLike, xvals: ArrayLike, yvals: ArrayLike, **kwargs
) -> np.ndarray:
    """
    Interpolate a set of values

    Parameters
    ----------
    x : ArrayLike, length n
        X values to interpolate at
    row : ArrayLike, shape (npdf, 1)
        Which rows to interpolate at
    xvals : ArrayLike, shape (npdf, npts)
        X-values used for the interpolation
    yvals : ArrayLike, shape (npdf, npts)
        Y-values used for the interpolation

    Returns
    -------
    vals : np.ndarray, shape (npdf, n)
        The interpolated values
    """
    rr = np.squeeze(row)
    nx = np.shape(x)[-1]

    def single_row(rv):
        return interp1d(xvals[rv], yvals[rv], **kwargs)(x)

    vv = np.vectorize(single_row, signature="()->(%i)" % (nx))
    return vv(rr)


def interpolate_multi_x_multi_y_2d(
    x: ArrayLike, row: ArrayLike, xvals: ArrayLike, yvals: ArrayLike, **kwargs
) -> np.ndarray:
    """
    Interpolate a set of values

    Parameters
    ----------
    x : ArrayLike, shape (npdf, n)
        X values to interpolate at
    row : ArrayLike, shape (npdf, 1)
        Which rows to interpolate at
    xvals : ArrayLike, shape (npdf, npts)
        X-values used for the interpolation
    yvals : ArrayLike, shape (npdf, npts)
        Y-values used for the interpolation

    Returns
    -------
    vals : np.ndarray, shape (npdf, n)
        The interpolated values
    """
    nx = np.shape(x)[-1]

    def evaluate_row(rv, xv):
        return interp1d(xvals[rv], yvals[rv], **kwargs)(xv)

    vv = np.vectorize(evaluate_row, signature="(),(%i)->(%i)" % (nx, nx))
    return vv(np.squeeze(row), x)


def interpolate_multi_x_multi_y(
    x: ArrayLike, row: ArrayLike, xvals: ArrayLike, yvals: ArrayLike, **kwargs
) -> np.ndarray:
    """
    Interpolate a set of values

    Parameters
    ----------
    x : ArrayLike, shape (npdf, n)
        X values to interpolate at
    row : ArrayLike, shape (npdf, 1)
        Which rows to interpolate at
    xvals : ArrayLike, shape (npdf, npts)
        X-values used for the interpolation
    yvals : ArrayLike, shape (npdf, npts)
        Y-values used for the interpolation

    Returns
    -------
    vals : np.ndarray
        The interpolated values
    """
    case_idx, xx, rr = get_eval_case(x, row)
    if case_idx in [CASE_PRODUCT, CASE_FACTOR]:
        return interpolate_multi_x_multi_y_product(xx, rr, xvals, yvals, **kwargs)
    if case_idx == CASE_2D:
        return interpolate_multi_x_multi_y_2d(xx, rr, xvals, yvals, **kwargs)
    return interpolate_multi_x_multi_y_flat(xx, rr, xvals, yvals, **kwargs)



from __future__ import annotations

from typing import Any
import numpy as np
from scipy import integrate as sciint
from scipy import interpolate as sciinterp

from ...utils.conversion import extract_xy_vals
from ..sparse_interp.sparse_rep import (
    build_sparse_representation,
    decode_sparse_indices,
)


# Conversion functions


def irreg_interp_extract_xy_vals(in_dist: "Ensemble", **kwargs):  # -> dict[str, Any]:
    """Wrapper for extract_xy_vals. Convert using a set of x and y values.

    Parameters
    ----------
    in_dist : Ensemble
        Input distributions
    xvals : np.ndarray[float]
        Locations at which the pdf is evaluated

    Returns
    -------
    data : dict
        The extracted data"""

    xvals = kwargs.pop("xvals", None)
    if xvals is None:  # pragma: no cover
        raise ValueError("To convert using extract_xy_vals you must specify xvals")
    return extract_xy_vals(in_dist, xvals)


def extract_vals_at_x(in_dist: "Ensemble", **kwargs) -> dict[str, np.ndarray[float]]:
    """Convert using a set of x and y values.

    Parameters
    ----------
    in_dist : Ensemble
        Input distributions

    Other Parameters
    ----------------
    xvals : np.ndarray[float]
        Locations at which the pdf is evaluated

    Returns
    -------
    data : dict[str, np.ndarray[float]]
        The extracted data
    """
    xvals = kwargs.pop("xvals", None)
    if xvals is None:  # pragma: no cover
        raise ValueError("To convert to extract_xy_vals you must specify xvals")
    yvals = in_dist.pdf(xvals)
    return dict(xvals=xvals, yvals=yvals)


def extract_xy_sparse(
    in_dist: "Ensemble", **kwargs
) -> dict[str, Any]:  # pragma: no cover
    """Extract xy-interpolated representation from an sparse representation

    Parameters
    ----------
    in_dist : Ensemble
        Input distributions

    Other Parameters
    ----------------
    xvals : ArrayLike
        Used to override the y-values
    xvals : ArrayLike
        Used to override the x-values
    nvals : int
        Used to override the number of bins

    Returns
    -------
    metadata : dict[str, Any]
        Dictionary with data for interpolated representation

    Notes
    -----
    This function will rebin to a grid more suited to the in_dist support by
    removing x-values corresponding to y=0
    """

    yvals = in_dist.objdata["yvals"]
    default = in_dist.metadata["xvals"][0]
    xvals = kwargs.pop("xvals", default)
    nvals = kwargs.pop("nvals", 300)
    # rebin to a grid more suited to the in_dist support
    xmin = np.min(xvals)
    _, j = np.where(yvals > 0)
    xmax = np.max(xvals[j])
    newx = np.linspace(xmin, xmax, nvals)
    interp = sciinterp.interp1d(xvals, yvals, assume_sorted=True)
    newpdf = interp(newx)
    sparse_indices, sparse_meta, A = build_sparse_representation(newx, newpdf)
    # decode the sparse indices into basis indices and weights
    basis_indices, weights = decode_sparse_indices(sparse_indices)
    # retrieve the weighted array of basis functions for each object
    pdf_y = A[:, basis_indices] * weights
    # normalize and sum the weighted pdfs
    x = sparse_meta["z"]
    y = pdf_y.sum(axis=-1)
    norms = sciint.trapezoid(y.T, x)
    y /= norms
    # super(sparse_gen, self).__init__(x, y.T, *args, **kwargs)
    xvals = x
    yvals = y.T
    return dict(xvals=xvals, yvals=yvals, **kwargs)
